from eflow._hidden.widgets.feature_data_cleaning_widget import *
from eflow._hidden.parent_objects.data_pipeline_segment import *
from eflow.foundation import DataFrameTypes
from eflow.utils.pandas_utils import check_if_feature_exists
from eflow.utils.misc_utils import get_parameters

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

import pandas as pd
import numpy as np
from scipy import stats
from collections import deque


class FeatureDataCleaner(DataPipelineSegment):
    """
    Designed for a multipurpose data cleaner.
    """

    def __init__(self,
                 segment_id=None,
                 create_file=True):
        """
        Args:
            segment_id: str
                Reference id to past segments of this object.

        Note/Caveats:
            When creating any public function that will be part of the pipeline's
            structure it is important to follow this given template. Also,
            try not to use _add_to_que. Can ruin the entire purpose of this
            project.
        """
        DataPipelineSegment.__init__(self,
                                     object_type=self.__class__.__name__,
                                     segment_id=segment_id,
                                     create_file=create_file)

        self.__test_cleaning_methods = False

        # self.__requires_nan_removal = df.isnull().values.any()
        #
        # # Throw error here
        # if df is None:
        #     return

        # --- Setting up widget options

        # Dummy line to show in the menu for cleaner viewing
        # self.__data_cleaning_options["TYPE"][
        #     "---------------------" + (" " * space_counters.pop())] = \
        #     self.__ignore_feature

        # Set up numerical cleaning options
        space_counters = {i for i in range(1, 50)}
        self.__data_cleaning_options = dict()
        self.__data_cleaning_options["Number"] = dict()

        self.__data_cleaning_options["Number"]["Ignore feature"] = \
            self.ignore_feature
        self.__data_cleaning_options["Number"]["Drop feature"] = \
            self.drop_feature
        self.__data_cleaning_options["Number"]["Remove all nans"] = \
            self.remove_nans

        self.__data_cleaning_options["Number"][
            "---------------------" + (" " * space_counters.pop())] = \
            self.ignore_feature

        self.__data_cleaning_options["Number"][
            "Fill nan with min value of distribution"] = \
            self.fill_nan_by_distribution
        self.__data_cleaning_options["Number"][
            "Fill nan with x% value of distribution"] = \
            self.fill_nan_by_distribution
        self.__data_cleaning_options["Number"][
            "Fill nan with median value of distribution"] = \
            self.fill_nan_by_distribution
        self.__data_cleaning_options["Number"][
            "Fill nan with max value of distribution"] = \
            self.fill_nan_by_distribution
        self.__data_cleaning_options["Number"][
            "---------------------" + (" " * space_counters.pop())] = \
            self.ignore_feature

        self.__data_cleaning_options["Number"][
            "Fill nan with average value of distribution"] = self.fill_nan_by_average
        self.__data_cleaning_options["Number"][
            "Fill nan with mode of distribution"] = self.fill_nan_by_mode
        self.__data_cleaning_options["Number"][
            "Fill null with specfic value"] = self.fill_nan_with_specfic_value
        self.__data_cleaning_options["Number"][
            "---------------------" + (" " * space_counters.pop())] = \
            self.ignore_feature

        self.__data_cleaning_options["Number"][
            "Fill with least common count of distribution"] = \
            self.fill_nan_by_occurance_percentaile
        self.__data_cleaning_options["Number"][
            "Fill with most common count of distribution"] = \
            self.fill_nan_by_occurance_percentaile
        self.__data_cleaning_options["Number"][
            "Fill with x% count distribution"] = \
            self.fill_nan_by_occurance_percentaile

        # Set up category cleaning options
        space_counters = {i for i in range(1, 50)}
        self.__data_cleaning_options["Category"] = dict()
        self.__data_cleaning_options["Category"]["Ignore feature"] = self.ignore_feature
        self.__data_cleaning_options["Category"]["Drop feature"] = \
            self.drop_feature
        self.__data_cleaning_options["Category"]["Remove all nans"] = \
            self.remove_nans

        self.__data_cleaning_options["Category"][
            "---------------------" + (" " * space_counters.pop())] = \
            self.ignore_feature
        self.__data_cleaning_options["Category"][
            "Fill null with specfic value"] = self.fill_nan_with_specfic_value
        self.__data_cleaning_options["Category"][
            "---------------------" + (" " * space_counters.pop())] = \
            self.ignore_feature

        self.__data_cleaning_options["Category"][
            "Fill nan with mode of distribution"] = self.fill_nan_by_mode
        self.__data_cleaning_options["Category"][
            "Fill with least common count of distribution"] = \
            self.fill_nan_by_occurance_percentaile
        self.__data_cleaning_options["Category"][
            "Fill with most common count of distribution"] = \
            self.fill_nan_by_occurance_percentaile
        self.__data_cleaning_options["Category"][
            "Fill with x% count distribution"] = \
            self.fill_nan_by_occurance_percentaile


        # Set up boolean cleaning options
        space_counters = {i for i in range(1, 50)}
        self.__data_cleaning_options["Bool"] = dict()
        self.__data_cleaning_options["Bool"][
            "Ignore feature"] = self.ignore_feature
        self.__data_cleaning_options["Bool"]["Drop feature"] = \
            self.drop_feature
        self.__data_cleaning_options["Bool"]["Remove all nans"] = \
            self.remove_nans

        self.__data_cleaning_options["Bool"][
            "---------------------" + (" " * space_counters.pop())] = \
            self.ignore_feature

        self.__data_cleaning_options["Bool"]["Make nan Assertions"] = self.make_nan_assertions

        self.__data_cleaning_options["Bool"][
            "Fill null with specfic bool value"] = self.fill_nan_with_specfic_value

        self.__data_cleaning_options["Bool"][
            "Fill with least common count of distribution"] = \
            self.fill_nan_by_occurance_percentaile
        self.__data_cleaning_options["Bool"][
            "Fill with most common count of distribution"] = \
            self.fill_nan_by_occurance_percentaile

        # Set up boolean cleaning options
        space_counters = {i for i in range(1, 50)}
        self.__data_cleaning_options["Date"] = dict()
        self.__data_cleaning_options["Date"][
            "Ignore feature"] = self.ignore_feature
        self.__data_cleaning_options["Date"]["Drop feature"] = \
            self.drop_feature

        # Error case on data types
        space_counters = {i for i in range(1, 50)}
        self.__data_cleaning_options["Unknown"] = dict()
        self.__data_cleaning_options["Unknown"][
            "ERROR UNKNOWN FEATURE TYPE FOUND"] = self.ignore_feature
        self.__data_cleaning_options["Unknown"][
            "Ignore feature"] = self.ignore_feature
        self.__data_cleaning_options["Unknown"]["Drop feature"] = \
            self.ignore_feature
        self.__data_cleaning_options["Unknown"]["Remove all nans"] = \
            self.remove_nans

        # Written conditionals for functions requiring input fields
        self.__require_input = {"Fill null with specfic value": None,
                                "Fill null with specfic bool value": 'x == 1 or x == 0',
                                "Fill nan with x% value of distribution":
                                    'x >= 0 and x <=100',
                                "Fill with random existing values": 'x > 0',
                                "Fill with x% count distribution":
                                    'x >= 0 and x <=100'}

        # ---
        # self.__notebook_mode = notebook_mode
        self.__ui_widget = DataCleaningWidget(
            require_input=self.__require_input,
            data_cleaning_options=self.__data_cleaning_options)


    def get_user_inputs(self):
        return self.__ui_widget.get_user_inputs()

    def run_widget(self,
                   df,
                   df_features,
                   nan_feature_names=[]):
        """
        df:
            A pandas dataframe object

        df_features:
            DataFrameTypes object; organizes feature types into groups.

        Returns:
            Returns a UI widget to create a JSON file for cleaning.
        """

        # Throw Error here
        if df is None:
            return

        if not nan_feature_names:
            nan_feature_names = df.columns[df.isna().any()].tolist()

        self.__ui_widget.run_widget(nan_feature_names,
                                    df_features)

    def perform_saved_widget_input(self,
                                   df,
                                   df_features,
                                   suppress_runtime_errors=True,
                                   reset_segment_file=False):

        try:
            for bool_value in [True,False]:
                self.__test_cleaning_methods = bool_value

                selected_options, \
                feature_input_holder, \
                feature_zscore_holder = self.__ui_widget.get_user_inputs()

                for feature_name, function_option in selected_options.items():

                    for dtype in ["Number","Bool","Category", "Date"]:

                        if function_option in self.__data_cleaning_options[dtype]:

                            saved_function = self.__data_cleaning_options[dtype][function_option]
                            exec_str = f"saved_function(df,df_features,feature_name,"

                            if feature_name in feature_input_holder and \
                                    feature_input_holder[feature_name]:
                                    try:
                                        int(feature_input_holder[feature_name])
                                        exec_str += feature_input_holder[
                                                        feature_name] + ","
                                    except:
                                        exec_str += "\"" + feature_input_holder[feature_name] + "\","

                            if function_option == "Fill nan with min value of distribution" or function_option == "Fill with least common count of distribution":
                                exec_str += "0,"
                            elif function_option == "Fill nan with median value of distribution":
                                exec_str += "50,"
                            elif function_option == "Fill nan with max value of distribution" or function_option == "Fill with most common count of distribution":
                                exec_str += "100,"

                            if feature_name in feature_zscore_holder and \
                                    feature_zscore_holder[feature_name]:
                                exec_str += feature_zscore_holder[feature_name] + ","

                            exec_str += ")"

                            if self.__test_cleaning_methods:
                                defined_function_string = exec_str.replace("saved_function",
                                                                           saved_function.__name__).replace("feature_name",
                                                                                                            f"\"{feature_name}\"")
                                print("***" * 10)
                                print("Testing function")
                                print(defined_function_string)
                            # print(exec_str)
                            try:
                                exec(exec_str)

                                if self.__test_cleaning_methods:
                                    print("PASSED TEST!")
                                    print("***" * 10)
                                print()
                            except Exception as e:

                                if self.__test_cleaning_methods:
                                    print("FAILED TEST!")
                                    print("***" * 10)
                                self.__test_cleaning_methods = False
                                if reset_segment_file:
                                    print("Exception hit when trying to perform all "
                                          "cleaning functions. "
                                          "Resetting json object for feature data cleaner segment!")
                                    self.reset_segment_file()
                                raise e
                            break
        finally:
            self.__test_cleaning_methods = False

    # --- Cleaning options
    def make_nan_assertions(self,
                            df,
                            df_features,
                            feature_name,
                            _add_to_que=True):
        """

            Make nan assertions for boolean features.

        Args:
            df: pd.Dataframe
                Pandas Dataframe

            df_features: DataFrameType from eflow
                Organizes feature types into groups.

            feature_name: string
                Name of the feature in the datatframe

            _add_to_que: bool
                Pushes the function to pipeline segment parent if set to 'True'.
        """

        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")


        if feature_name not in df_features.bool_features():
            raise UnsatisfiedRequirments(f"{feature_name} must be a bool feature.")

        unique_series = df[feature_name].dropna().unique().tolist()

        if len(unique_series) == 1 and (
                unique_series[0] == 1 or unique_series[0] == 0):
            replace_value = int(unique_series[0] == 1)

            self.fill_nan_with_specfic_value(df,
                                             df_features,
                                             feature_name=feature_name,
                                             replace_value=replace_value,
                                             _add_to_que=_add_to_que)

        else:
            raise UnsatisfiedRequirments(f"Boolean assertions can't be made with this given feature {feature_name}.")

    def ignore_feature(self,
                       df,
                       df_features,
                       feature_name,
                       _add_to_que=True):
        """
            Ignore the given feature.

        Args:
            df: pd.Dataframe
                Pandas Dataframe

            df_features: DataFrameType from eflow
                Organizes feature types into groups.

            feature_name: string
                Name of the feature in the datatframe

            _add_to_que: bool
                Pushes the function to pipeline segment parent if set to 'True'.
        """
        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        if not self.__test_cleaning_methods:
            print("Ignore feature: ", feature_name)

            if _add_to_que:
                params_dict = locals()
                parameters = get_parameters(self.drop_feature)

                self._DataPipelineSegment__add_function_to_que("ignore_feature",
                                                               parameters,
                                                               params_dict)

            # Remove any unwanted arguments in params_dict
            # if _add_to_que:
            #     params_dict = locals()
            #     parameters = get_parameters(self.ignore_feature)
            #
            #     self._DataPipelineSegmentfunction_name("ignore_feature",
            #                                            parameters,
            #                                            params_dict)


    def drop_feature(self,
                     df,
                     df_features,
                     feature_name,
                     _add_to_que=True):
        """
            Drop a feature in the dataframe.

        Args:
            df: pd.Dataframe
                Pandas Dataframe

            df_features: DataFrameType from eflow
                Organizes feature types into groups.

            feature_name: string
                Name of the feature in the datatframe

            _add_to_que: bool
                Pushes the function to pipeline segment parent if set to 'True'.
        """
        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        if not self.__test_cleaning_methods:

            print("Droping Feature: ", feature_name)

            df.drop(columns=feature_name,
                    inplace=True)
            df.reset_index(drop=True,
                           inplace=True)
            df_features.remove_feature(feature_name)

            if _add_to_que:
                params_dict = locals()
                parameters = get_parameters(self.drop_feature)

                self._DataPipelineSegment__add_function_to_que("drop_feature",
                                                               parameters,
                                                               params_dict)

    def remove_nans(self,
                    df,
                    df_features,
                    feature_name,
                    _add_to_que=True):
        """

            Remove rows of data based on the given feature.

        Args:
            df: pd.Dataframe
                Pandas Dataframe

            df_features: DataFrameType from eflow
                Organizes feature types into groups.

            feature_name: string
                Name of the feature in the datatframe

            _add_to_que: bool
                Pushes the function to pipeline segment parent if set to 'True'.
        """
        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        if not self.__test_cleaning_methods:

            print(f"Remove data from rows where the feature {feature_name} is equal to nan")

            df[feature_name].dropna(inplace=True)
            df.reset_index(drop=True,
                           inplace=True)
            df_features.remove_feature(feature_name)

            if _add_to_que:

                # Remove any unwanted arguments in params_dict
                params_dict = locals()
                parameters = get_parameters(self.remove_nans)

                self._DataPipelineSegment__add_function_to_que("remove_nans",
                                                               parameters,
                                                               params_dict)


    def fill_nan_by_distribution(self,
                                 df,
                                 df_features,
                                 feature_name,
                                 percentile,
                                 z_score=None,
                                 _add_to_que=True):
        """

            Fill nan by the distribution of data.

        Args:
            df: pd.Dataframe
                Pandas Dataframe

            df_features: DataFrameType from eflow
                Organizes feature types into groups.

            feature_name: string
                Name of the feature in the datatframe

            percentile: float or int

            z_score:

            _add_to_que: bool
                Pushes the function to pipeline segment parent if set to 'True'.
        """
        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        if feature_name in df_features.continuous_numerical_features():
            series_obj = df[feature_name].sort_values()
        else:
            series_obj = df.sort_values([feature_name],
                                        ascending=True).groupby(feature_name).head(float("inf"))[feature_name]

        if z_score:
            if isinstance(z_score, float) or isinstance(z_score, int):
                series_obj = self.__zcore_remove_outliers(series_obj.to_frame(),
                                                          feature_name,
                                                          z_score)
            else:
                raise ValueError("Z-Score must be at numerical value.")
        else:
            series_obj = df[feature_name].dropna()

        replace_value = np.percentile(series_obj, percentile)

        # Remove any unwanted arguments in params_dict

        if not self.__test_cleaning_methods:
            print(f"Fill nan on distribution; {percentile}% of {feature_name}")
            self.fill_nan_with_specfic_value(df,
                                             df_features,
                                             feature_name=feature_name,
                                             replace_value=replace_value,
                                             _add_to_que=_add_to_que)

    def fill_nan_by_average(self,
                            df,
                            df_features,
                            feature_name,
                            z_score=None,
                            _add_to_que=True):

        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        params_dict = locals()

        # Remove any unwanted arguments in params_dict
        if _add_to_que:
            params_dict = locals()
            for arg in ["self", "df", "df_features", "_add_to_que",
                        "params_dict"]:
                del params_dict[arg]

        if feature_name not in df_features.continuous_numerical_features():
            raise UnsatisfiedRequirments(f"{feature_name} must be a saved as float or integer in df_features")

        if z_score:
            if isinstance(z_score,float) or isinstance(z_score,int):
                series_obj = self.__zcore_remove_outliers(df,
                                                          feature_name,
                                                          z_score)
            else:
                raise ValueError("Z-Score must be at numerical value.")
        else:
            series_obj = df[feature_name].dropna()

        replace_value = series_obj.mean()

        if not self.__test_cleaning_methods:
            print("Fill nan based on the average of the distribution.")
            self.fill_nan_with_specfic_value(df,
                                             df_features,
                                             feature_name=feature_name,
                                             replace_value=replace_value,
                                             _add_to_que=_add_to_que)

    def fill_nan_by_mode(self,
                         df,
                         df_features,
                         feature_name,
                         z_score=None,
                         _add_to_que=True):

        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        if z_score:
            series_obj = self.__zcore_remove_outliers(df,
                                                      feature_name,
                                                      z_score)
        else:
            series_obj = df[feature_name].dropna()

        mode_series = series_obj.mode()
        if not len(mode_series):
            pass
        else:
            replace_value = mode_series[0]

        if not self.__test_cleaning_methods:
            print("Fill nan by mode")
            self.fill_nan_with_specfic_value(df,
                                             df_features,
                                             feature_name=feature_name,
                                             replace_value=replace_value,
                                             _add_to_que=_add_to_que)

    def fill_nan_with_specfic_value(self,
                                    df,
                                    df_features,
                                    feature_name,
                                    replace_value,
                                    _add_to_que=True):

        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        try:
            replace_value = replace_value.item()
        except AttributeError:
            pass

        if not self.__test_cleaning_methods:
            df[feature_name].fillna(replace_value,
                                    inplace=True)

            if _add_to_que:
                params_dict = locals()
                parameters = get_parameters(self.fill_nan_with_specfic_value)

                print("Replace nan with {0} on feature: {1}".format(replace_value,
                                                                    feature_name))

                self._DataPipelineSegment__add_function_to_que("fill_nan_with_specfic_value",
                                                               parameters,
                                                               params_dict)


    def fill_nan_by_occurance_percentaile(self,
                                          df,
                                          df_features,
                                          feature_name,
                                          percentaile,
                                          z_score=None,
                                          _add_to_que=True):
        check_if_feature_exists(df,
                                feature_name)

        if feature_name not in df_features.all_features():
            raise KeyError(
                f"The feature \'{feature_name}\' was not found in the dataframe!"
                + " Please select a valid feature from the df_features.")

        # Remove any unwanted arguments in params_dict
        if _add_to_que:
            params_dict = locals()
            for arg in ["self", "df", "df_features", "_add_to_que",
                        "params_dict"]:
                try:
                    del params_dict[arg]
                except KeyError:
                    pass

        if z_score:
            series_obj = self.__zcore_remove_outliers(df,
                                                      feature_name,
                                                      z_score)
        else:
            series_obj = df[feature_name].dropna()

        array = np.asarray(series_obj.value_counts() / df.dropna().shape[0])
        idx = (np.abs(array - (percentaile / 100))).argmin()
        replace_value = series_obj.value_counts().keys()[idx]

        if not self.__test_cleaning_methods:
            print("Fill nan by occurance percentaile.")
            self.fill_nan_with_specfic_value(df,
                                             df_features,
                                             feature_name=feature_name,
                                             replace_value=replace_value,
                                             _add_to_que=_add_to_que)

    def __zcore_remove_outliers(self,
                                df,
                                feature_name,
                                zscore_val,
                                _add_to_que=True):

        z_score_return = 2

        print("After the zscore applied of {0} to -{0}".format(zscore_val))

        return df[feature_name].dropna()[
            (z_score_return >= (zscore_val * -1)) & (
                    z_score_return <= zscore_val)]





from sklearn.model_selection import GridSearchCV

from eflow.utils.math_utils import euclidean_distance

import numpy as np
from scipy import stats

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


def get_cluster_probas(center_points,
                       data):

    for dp in data:

        cluster_distances = []
        for cp in center_points:
            cluster_distances.append(euclidean_distance(cp, dp))

        probas = 1 - np.array(cluster_distances) / (
            np.array(cluster_distances).sum())
        return probas / probas.sum()


def find_all_zscore_distances_from_target(data_points,
                                          target_point):
    """
        Finds all distances between the target and the other points.
    """
    distances = []
    for dp in data_points:
        distances.append(euclidean_distance(dp, target_point))

    return stats.zscore(np.array(distances))


def optimize_model_grid(model,
                        X_train,
                        y_train,
                        param_grid,
                        scoring="f1_macro",
                        cv=5,
                        verbose=0,
                        n_jobs=1):
    """

        Finds the best parameters for a model; returns the model and parameters.

    Args:
        model:
            Machine learning model to fit across a cross fold with a cross

        X_train:
            Feature matrix.

        y_train:
            Target vector.

        param_grid:
            Dictionary with parameters names.

        scoring:
            String value to determine the metric to evaluate the best model.
            Link to all strings: http://tinyurl.com/y22f3m5k

        cv:
            Cross-validation strategy for 'training'.

        verbose:
            Controls the verbosity: the higher, the more messages.

        n_jobs:
            Number of jobs to run in parallel.

        Note:
            From the sklearn library;

    Returns:
        Finds the best parameters for a grid; returns the model and parameters.
        Link: http://tinyurl.com/y296u6mq
    """

    # Instantiate the GridSearchCV object
    model_cv = GridSearchCV(model,
                            param_grid,
                            cv=cv,
                            n_jobs=n_jobs,
                            verbose=verbose,
                            scoring=scoring)

    # Fit it to the data
    model_cv.fit(X_train, y_train)

    # Print the tuned parameters and score
    print("Tuned Parameters: {}".format(model_cv.best_params_))
    print("Best score on trained data was {0:4f}".format(model_cv.best_score_))

    # Return model and parameters
    return model_cv.best_estimator_, model_cv.best_params_




from eflow._hidden.parent_objects import DataPipelineSegment
from eflow._hidden.constants import BOOL_STRINGS
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments
from eflow.utils.language_processing_utils import get_synonyms
from eflow.utils.misc_utils import get_parameters

import copy
import pandas as pd
import numpy as np

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class DataEncoder(DataPipelineSegment):
    """
        Attempts to convert features to the correct types. Will update the
        dataframe and df_features.
    """
    def __init__(self,
                 segment_id=None,
                 create_file=True):
        """
        Args:
            segment_id:
                Reference id to past segments of this object.

        Note/Caveats:
            When creating any public function that will be part of the pipeline's
            structure it is important to follow this given template. Also,
            try not to use _add_to_que. Can ruin the entire purpose of this
            project.
        """
        DataPipelineSegment.__init__(self,
                                     object_type=self.__class__.__name__,
                                     segment_id=segment_id,
                                     create_file=create_file)

    def encode_data(self,
                    df,
                    df_features,
                    apply_value_representation=True,
                    _add_to_que=True):
        """

            Encode the data into numerical values for machine learning processes.

        Args:
            df: pd.Dataframe
                Pandas dataframe.

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.

            apply_value_representation: bool
                Translate features into most understandable/best representation/

            _add_to_que: bool
                Hidden variable to determine if the function should be pushed
                to the pipeline segment.
        """

        # Apply value representation to feature values
        if apply_value_representation:
            feature_value_represention = df_features.get_feature_value_representation()

            # Inverse dict
            tmp_dict = copy.deepcopy(feature_value_represention)
            for feature_name in feature_value_represention.keys():
                tmp_dict[feature_name] = dict()
                for val, reprs in feature_value_represention[feature_name].items():
                    tmp_dict[feature_name][reprs] = val

            feature_value_represention = tmp_dict

            for feature_name in feature_value_represention.keys():
                if feature_name not in df.columns:
                    continue

                if df[feature_name].dtype == "O":
                    df[feature_name].replace(
                        feature_value_represention[feature_name],
                        inplace=True)

        # Decode data from categorical values to proper strings.
        encoder_dict = df_features.get_label_encoder()
        for feature_name in encoder_dict.keys():

            if feature_name not in df.columns:
                continue

            if df[feature_name].dtype == "O":
                df[feature_name].replace(encoder_dict[feature_name],
                                         inplace=True)

                df_features.set_feature_to_categorical(feature_name)

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.encode_data)
            self._DataPipelineSegment__add_function_to_que("encode_data",
                                                           parameters,
                                                           params_dict)

    def decode_data(self,
                    df,
                    df_features,
                    apply_value_representation=True,
                    _add_to_que=True):
        """

            Decode the data into non-numerical values for more descriptive analysis.

        Args:
            df: pd.Dataframe
                Pandas dataframe.

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.

            apply_value_representation: bool
                Translate features into most understandable/best representation/

            _add_to_que: bool
                Hidden variable to determine if the function should be pushed
                to the pipeline segment.
        """
        # Decode data from categorical values to proper strings.
        decoder_dict = df_features.get_label_decoder()
        for feature_name in decoder_dict.keys():

            if feature_name not in df.columns:
                continue

            if df[feature_name].dtype != "O":
                df[feature_name].replace(decoder_dict[feature_name],
                                         inplace=True)

        # Apply value representation to feature values
        if apply_value_representation:
            feature_value_represention = df_features.get_feature_value_representation()
            # Replace values by each corresponding feature value related dict
            for feature_name in feature_value_represention.keys():
                if feature_name not in df.columns:
                    continue

                if df[feature_name].dtype == "O":
                    df[feature_name].replace(feature_value_represention[feature_name],
                                             inplace=True)

                df_features.set_feature_to_string(feature_name)

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.decode_data)
            self._DataPipelineSegment__add_function_to_que("decode_data",
                                                           parameters,
                                                           params_dict)

    def apply_binning(self,
                      df,
                      df_features,
                      binable_features=[],
                      _add_to_que=True):

        # Remove any unwanted arguments in params_dict
        params_dict = locals()
        for arg in ["self", "df", "df_features", "_add_to_que",
                    "params_dict"]:
            try:
                del params_dict[arg]
            except KeyError:
                pass

        if not binable_features:
            binable_features = df.columns

        # Apply binning
        for feature_name in binable_features:
            bin_labels_dict = df_features.get_feature_binning(feature_name)
            if bin_labels_dict:
                # Convert to category data
                df[feature_name] = pd.to_numeric(df[feature_name].dropna(),
                                                 errors='coerce')
                df[feature_name] = pd.cut(df[feature_name],
                                          bins=bin_labels_dict["bins"],
                                          labels=bin_labels_dict["labels"])

                # Feature set to categorical
                df_features.set_feature_to_categorical(feature_name)

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.decode_data)
            self._DataPipelineSegment__add_function_to_que("apply_binning",
                                                           parameters,
                                                           params_dict)

    def apply_value_representation(self,
                                   df,
                                   df_features,
                                   _add_to_que=True):
        """

            Translate features into most understandable/best representation

        Args:
            df: pd.Dataframe
                Pandas dataframe.

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.

            _add_to_que: bool
                Hidden variable to determine if the function should be pushed
                to the pipeline segment.
        """
        feature_value_represention = df_features.get_feature_value_representation()

        # Replace values by each corresponding feature value related dict
        for feature_name in feature_value_represention:

            if feature_name not in df.columns:
                raise KeyError(
                    f"Dataframe doesn't have feature name '{feature_name}'.")

            df[feature_name].replace(feature_value_represention[feature_name],
                                     inplace=True)

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.decode_data)
            self._DataPipelineSegment__add_function_to_que("apply_value_representation",
                                                           parameters,
                                                           params_dict)

    def revert_value_representation(self,
                                    df,
                                    df_features,
                                    _add_to_que=True):
        """

            Translate features back into worst representation

        Args:
            df: pd.Dataframe
                Pandas dataframe.

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.

            _add_to_que: bool
                Hidden variable to determine if the function should be pushed
                to the pipeline segment.
        """
        feature_value_represention = df_features.get_feature_value_representation()

        # Replace values by each corresponding feature value related dict
        for feature_name in feature_value_represention:

            if feature_name not in df.columns:
                raise KeyError(
                    f"Dataframe doesn't have feature name '{feature_name}'.")


            df[feature_name].replace({v: k
                                      for k, v in feature_value_represention[
                                          feature_name].items()},
                                     inplace=True)

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.revert_value_representation)
            self._DataPipelineSegment__add_function_to_que("revert_value_representation",
                                                           parameters,
                                                           params_dict)

    def make_values_bool(self,
                         df,
                         df_features,
                         _add_to_que=True):
        """

            Convert all string bools to numeric bool value
        Args:
            df: pd.Dataframe
                Pandas dataframe.

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.

            _add_to_que: bool
                Hidden variable to determine if the function should be pushed
                to the pipeline segment.
        """
        for bool_feature in df_features.bool_features():
            if df[bool_feature].dtype == "O":
                bool_check,true_val,false_val = self.__bool_string_values_check(
                    df[bool_feature].dropna().unique())

                # Replace bool string values with bools
                if bool_check:
                    df[bool_feature].replace({true_val:1,
                                              false_val:0},
                                             inplace=True)
        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.make_values_bool)
            self._DataPipelineSegment__add_function_to_que("make_values_bool",
                                                           parameters,
                                                           params_dict)

    def make_dummies(self,
                     df,
                     df_features,
                     qualitative_features=[],
                     _feature_values_dict=None,
                     _add_to_que=True):
        """

            Create dummies features of based on qualtative feature data and removes
            the original feature.

            Note
                _feature_values_dict does not need to be init. Used for backend
                resource.

        Args:
            df: pd.Dataframe
                Pandas dataframe.

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.

            qualtative_features: collection of strings
                Feature names to convert the feature data into dummy features.

            _add_to_que: bool
                Hidden variable to determine if the function should be pushed
                to the pipeline segment.
        """
        # Convert to the correct types
        if isinstance(qualitative_features,str):
            qualtative_features = [qualitative_features]

        if not _feature_values_dict:
            _feature_values_dict = dict()

        pd.set_option('mode.chained_assignment', None)

        for cat_feature in qualitative_features:

            if cat_feature not in df_features.string_features() | df_features.categorical_features():
                raise UnsatisfiedRequirments(f"No feature named '{cat_feature}' in categorical or string features.")

            if cat_feature not in _feature_values_dict:
                _feature_values_dict[cat_feature] = df[cat_feature].dropna().unique()
                _feature_values_dict[cat_feature].sort()
                _feature_values_dict[cat_feature] = _feature_values_dict[cat_feature].tolist()

            dummy_features = []
            for feature_value in  _feature_values_dict[cat_feature]:
                new_feature = cat_feature + f"_{feature_value}"
                bool_array = df[cat_feature] == feature_value
                df[new_feature] = copy.deepcopy(bool_array)
                dummy_features.append(new_feature)

            # # Make dummies and remove original feature
            # dummies_df = pd.get_dummies(_feature_values_dict[cat_feature],
            #                             prefix=cat_feature)

            df.drop(columns=[cat_feature],
                    inplace=True)
            df_features.remove_feature(cat_feature)
            df_features.set_feature_to_dummy_encoded(cat_feature,
                                                     dummy_features)

            # # Apply to dataframe
            # for feature_name in dummies_df.columns:
            #     df[feature_name] = dummies_df[feature_name]

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.make_dummies)
            self._DataPipelineSegment__add_function_to_que("make_dummies",
                                                           parameters,
                                                           params_dict)

    def revert_dummies(self,
                       df,
                       df_features,
                       qualitative_features=[],
                       _add_to_que=True):
        """

            Convert dummies features back to the original feature.

        Args:
            df: pd.Dataframe
                Pandas dataframe.

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.

            qualitative_features: collection of strings
                Feature names to convert the dummy features into original feature
                data.

            _add_to_que: bool
                Hidden variable to determine if the function should be pushed
                to the pipeline segment.
        """

        df.reset_index(inplace=True,
                       drop=True)

        if isinstance(qualitative_features, str):
            feature_name = [qualitative_features]

        for feature_name in qualitative_features:
            dummies_df = df[
                df_features.get_dummy_encoded_features()[feature_name]]
            dummies_columns = dummies_df.columns.to_list()

            tmp_df = dummies_df[dummies_df == 1].stack().reset_index()
            del dummies_df

            df[feature_name] = np.full([len(df)], np.nan)
            df[feature_name].iloc[tmp_df["level_0"]] = tmp_df[
                "level_1"].values.tolist()

            # Remove dummy features
            df.drop(columns=dummies_columns,
                    inplace=True)

            df[feature_name] = df[feature_name].str[len(feature_name) + 1:]

            # Remove dummy encoded relationship
            df_features.remove_feature_from_dummy_encoded(feature_name)

            # Add feature back to original set in df_features
            try:
                pd.to_numeric(df[feature_name].dropna())
                df_features.add_new_categorical_feature(feature_name)

            except ValueError:
                df_features.add_new_string_feature(feature_name)

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.revert_dummies)
            self._DataPipelineSegment__add_function_to_que("revert_dummies",
                                                           parameters,
                                                           params_dict)


    def __bool_string_values_check(self,
                                   feature_values):
        """

            Checks if a collection of strings can be considered a bool feature
            based on the amount of strings and the values of those strings.

            Note -
                Modified from data frame types
        Args:
            feature_values: collection
                Collection of strings to apply natural language process to
                determine if the series data is boolean or not.

        Returns:
            Returns true or false if the values can be considered a bool and
            the true and false values found.
        """

        if len(feature_values) > 2:
            return False, None, None

        found_true_value = None
        found_false_value = None

        for val in feature_values:

            if not isinstance(val,str):
                continue

            org_val = copy.deepcopy(val)
            val = val.lower()

            # Determine if val is true
            if not found_true_value:

                # Check if the string already exist in the defined set
                if val in BOOL_STRINGS.TRUE_STRINGS:
                    found_true_value = org_val
                    continue
                else:
                    # Attempt to find synonyms of the defined words to compare to
                    # the iterable string
                    for true_string in BOOL_STRINGS.TRUE_STRINGS:

                        if len(true_string) < 2:
                            continue

                        for syn in get_synonyms(true_string):
                            if syn == val:
                                found_true_value = org_val
                                continue

            # -----
            if not found_false_value:

                # -----
                if val in BOOL_STRINGS.FALSE_STRINGS:
                    found_false_value = org_val
                    continue
                else:
                    # -----
                    for false_string in BOOL_STRINGS.FALSE_STRINGS:

                        if len(false_string) < 2:
                            continue

                        for syn in get_synonyms(false_string):
                            if syn == val:
                                found_false_value = org_val
                                continue

        if len(feature_values) == 2:
            return isinstance(found_true_value,str) and isinstance(found_false_value,str), found_true_value, found_false_value
        else:
            return isinstance(found_true_value,str) or isinstance(found_false_value,str), found_true_value, found_false_value



from eflow.foundation import DataFrameTypes
from eflow.utils.sys_utils import create_dir_structure, check_if_directory_exists, \
    correct_directory_path, get_all_files_from_path, get_unique_directory_path, \
    get_all_directories_from_path, json_file_to_dict
import os
from eflow._hidden.constants import SYS_CONSTANTS
import shutil
import copy
from eflow.utils.language_processing_utils import get_synonyms


def move_folder_to_eflow_garbage(directory_path,
                                 create_sub_dir=None):
    """

        Renames and moves contents to a folder labeled 'Garbage' for the user/system
        to later handle.

    Args:
        directory_path:
            Path to given folder to move to 'Garbage'

        create_sub_dir:
            If the folder 'Garbage' needs further organization then you can specify
            a folder for the given folder to be embedded in.
    """
    directory_path = correct_directory_path(directory_path)
    check_if_directory_exists(directory_path)

    if not create_sub_dir:
        create_sub_dir = ""
    else:
        correct_directory_path(create_sub_dir)

    garbage_folder_path = create_dir_structure(os.getcwd(),
                                               f"{SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME}/_Extras/Garbage/{create_sub_dir}")

    path_to_folder, folder_name = directory_path[:-1].rsplit('/', 1)

    _, folder_name = get_unique_directory_path(garbage_folder_path,folder_name).rsplit('/', 1)

    os.rename(directory_path,
              f'{path_to_folder}/{folder_name}')

    shutil.move(f'{path_to_folder}/{folder_name}', garbage_folder_path)



def get_type_holder_from_pipeline(pipeline_name):
    """

        Returns a type holder object from a specfied pipeline directory name.

    Args:
        pipeline_name:
            The name of the pipeline to json file to init type holder.

    Returns:
        Init type holder.
    """
    directory_path = correct_directory_path(f"{os.getcwd()}/{SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME}/_Extras/Pipeline Structure/Data Pipeline/{pipeline_name}")
    df_features = DataFrameTypes(df=None)
    directory_path = correct_directory_path(directory_path)
    df_features.init_on_json_file(directory_path + "df_features.json")
    return df_features


def create_color_dict_for_features(df,
                                   df_features,
                                   value_limit=50):
    """

        Attempts to give a color to each value based on the feature name and
        the feature value. This should be used as a template for the user to insert
        their own hex values.

    Note:
        Easily the messiest code I have ever written but it gets the job done...
        I am so sorry for your eyes...
    """


    '''
    First row of matrix determines the feature's name. Empty space represents
    Second row of matrix determines the value inside the feature and the associated color to that value.
    '''
    defined_feature_value_colors = list()

    defined_feature_value_colors.append([["Died", "Dead", "Deceased"],
                                         ["1", "True",
                                          "#616369"],
                                         ["0", "False",
                                          "#4dad6c"]])

    defined_feature_value_colors.append([["Survived", "Lived"],
                                         ["1", "true",
                                          "#4dad6c"],
                                         ["0", "false",
                                          "#616369"]])

    defined_feature_value_colors.append([["gender", "sex",],
                                         ["Male", "M",
                                          "#7EAED3"],
                                         ["Female", "F",
                                          "#FFB6C1"]])

    defined_feature_value_colors.append([[" "],
                                         ["y" "yes",
                                          "#55a868"],
                                         ["n", "no",
                                          "#ff8585"]])
    defined_feature_value_colors.append([[" "],
                                         ["1", "true",
                                          "#55a868"],
                                         ["0", "false",
                                          "#ff8585"]])

    all_true_vals = {"1","y","yes","true","t","y"}
    all_false_vals = {"0","n","no","false","f","n"}

    feature_value_color_dict = dict()

    # -----
    for feature_name in df_features.non_continuous_features():

        feature_values = []
        feature_value_color_dict[feature_name] = dict()
        for value in set(df[feature_name].dropna().values):
            feature_value_color_dict[feature_name][str(value)] = None
            feature_values.append(str(value))


        if len(feature_values) >= value_limit:
            del feature_value_color_dict[feature_name]
            feature_value_color_dict[feature_name] = dict()
            continue


        # Check if the given feature name matches any pre-defined names
        for defined_feature_info in copy.deepcopy(defined_feature_value_colors):

            # Extract all pre-defined feature names
            defined_feature_names = [str(x).lower()
                                     for x in defined_feature_info.pop(0)]

            # Check if any of the pre-defined feature names have synoymns with the given feature name
            feature_synonym_found = False
            for given_feature_name in defined_feature_names:
                synonyms = get_synonyms(given_feature_name)

                if feature_name in synonyms:
                    feature_synonym_found = True

            # Compare both feature names; ignore char case; check for default
            if feature_synonym_found or \
                    feature_name.lower() in defined_feature_names or \
                    " " in defined_feature_names:
                defined_feature_values = [[j.lower() for j in x]
                                          for x in defined_feature_info]

                # Finally make logical color assertions based on the feature values
                found_colors = False
                for defined_values in defined_feature_values:

                    # Check if the feature value's synonym's with pre defined feature values
                    val_synonym_found = False
                    defined_color = defined_values.pop(-1)
                    for feature_val in feature_values:

                        for _defined_val in defined_values:
                            synonyms = get_synonyms(str(_defined_val))

                            if feature_name in synonyms:
                                val_synonym_found = True

                        # -----
                        if feature_val.lower() in defined_values or val_synonym_found:
                            if feature_name in df_features.bool_features():

                                if feature_val in all_true_vals:
                                    feature_value_color_dict[feature_name][
                                        "1"] = defined_color

                                elif feature_val in all_false_vals:
                                    feature_value_color_dict[feature_name][
                                        "0"] = defined_color
                            else:
                                feature_value_color_dict[feature_name][
                                    feature_val] = defined_color

                            found_colors = True

                # Break pre-defined feature names loop
                if found_colors:
                    break

    tmp_feature_value_color_dict = copy.deepcopy(feature_value_color_dict)

    # Convert any feature values back to numeric
    for feature_name,color_value_dict in tmp_feature_value_color_dict.items():
        for feature_val, color in color_value_dict.items():

            try:
                int(feature_val)

                del feature_value_color_dict[feature_name][feature_val]

                feature_value_color_dict[feature_name][int(feature_val)] = color

            except:
                pass

    return feature_value_color_dict


def remove_unconnected_pipeline_segments():
    """

        Removes all pipeline segments that aren't connected to a pipeline structure.
    """


    pipeline_struct_dir = os.getcwd() + f"/{SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME}/_Extras/Pipeline Structure/"

    if not os.path.exists(pipeline_struct_dir):
        print("Project structure for pipelines has yet to be initalized. Can't clean/remove any files related to pipeline...")
    else:
        segment_dict = dict()
        pipeline_segments_dict = dict()
        # Get all segment files by their types.
        if os.path.exists(pipeline_struct_dir + "/Data Pipeline Segments/"):
            all_segment_dirs = get_all_directories_from_path(
                pipeline_struct_dir + "/Data Pipeline Segments")

            for segment_type in all_segment_dirs:
                segment_dict[segment_type] = get_all_files_from_path(
                    pipeline_struct_dir + f"/Data Pipeline Segments/{segment_type}")

        # Get all segments related to each pipeline.
        if os.path.exists(pipeline_struct_dir + "/Data Pipeline/"):
            all_pipeline_dirs = get_all_directories_from_path(
                pipeline_struct_dir + "/Data Pipeline/")

            for pipeline_name in all_pipeline_dirs:
                json_file = json_file_to_dict(
                    f"{pipeline_struct_dir}/Data Pipeline/{pipeline_name}/root_pipeline.json")

                for i in range(1, json_file["Pipeline Segment Count"] + 1):
                    segment_id = json_file["Pipeline Segment Order"][str(i)][
                        'Pipeline Segment ID']
                    segment_type = json_file["Pipeline Segment Order"][str(i)][
                        'Pipeline Segment Type']

                    if segment_type in segment_dict.keys() and segment_id + ".json" in \
                            segment_dict[segment_type]:
                        segment_dict[segment_type].remove(segment_id + ".json")

        # Create path to eflow's garbage
        garbage_folder_path = create_dir_structure(os.getcwd(),
                                                   f"{SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME}/_Extras/Garbage/Data Pipeline Segments/DataTransformer/")

        # Rename files and move them to appropriate
        for segment_type, segment_ids in segment_dict.items():
            files_in_garbage = get_all_files_from_path(garbage_folder_path)

            for _id in segment_ids:
                file_to_remove = _id
                i = 1
                while file_to_remove in files_in_garbage:
                    file_to_remove = _id
                    file_to_remove = file_to_remove.split(".")[
                                         0] + f"_{i}.json"
                    i += 1

                os.rename(
                    pipeline_struct_dir + f"Data Pipeline Segments/{segment_type}/{_id}",
                    pipeline_struct_dir + f"Data Pipeline Segments/{segment_type}/{file_to_remove}")
                shutil.move(
                    pipeline_struct_dir + f"Data Pipeline Segments/{segment_type}/{file_to_remove}",
                    garbage_folder_path + file_to_remove)


from .file_output import FileOutput
from .data_pipeline_segment import DataPipelineSegment
from .pipeline_jupyter_widget import PipelineJupyterWidget
from .data_analysis import DataAnalysis
from .model_analysis import ModelAnalysis
from .auto_modeler import AutoModeler


from eflow._hidden.custom_exceptions import UnsatisfiedRequirments
from eflow._hidden.constants import SYS_CONSTANTS
from eflow.utils.sys_utils import create_dir_structure
from eflow.utils.string_utils import correct_directory_path
import os
import copy

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class FileOutput(object):
    """
        Ensures a folder is part of eflow's main output stream and creates a
        sub directory based on the arg project_name.
        Note:
              If 'project_name' string is formatted in a relative path form
              then it will generate the provided directory.

              Ex:
                Dir A/Best Project. Will generate folders "Dir A" and
                Best Project.
    """

    def __init__(self,
                 project_name,
                 overwrite_full_path=None):
        """
        Args:
            project_name: string
                Sub directory to create on top of the directory
                'PARENT_OUTPUT_FOLDER_NAME'.

            overwrite_full_path: string
                The passed directory path must already exist. Will completely
                ignore the project name and attempt to point to this already
                created directory.
        """

        # Setup project structure
        if not overwrite_full_path:
            parent_structure = "/" + SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME \
                               + "/" + project_name + "/"

            create_dir_structure(os.getcwd(),
                                       parent_structure)
            tmp_path = correct_directory_path(
                os.getcwd() + parent_structure)

        # Trusting the user that this path must already exist
        else:
            overwrite_full_path = correct_directory_path(overwrite_full_path)

            # Path doesn't contain eflow's main output
            if f"/{SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME}/" not in overwrite_full_path:
                raise UnsatisfiedRequirments(f"Directory path must have {SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME} "
                                             f"as a directory name or this program will not work correctly.")

            # Unknown path found
            if not os.path.exists(overwrite_full_path):
                raise SystemError("The path must already be defined in full on "
                                  "your system to use a different directory "
                                  "structure than orginally intended.")

            tmp_path = overwrite_full_path

        from eflow._hidden.general_objects import enum
        self.__PROJECT = enum(PATH_TO_OUTPUT_FOLDER=tmp_path,
                              RELATIVE_PATH_TO_OUTPUT_FOLDER=tmp_path.split(f"/{SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME}/")[1])

    @property
    def folder_path(self):
        """

            Path to folder
        """
        return copy.deepcopy(self.__PROJECT.PATH_TO_OUTPUT_FOLDER)

    @property
    def relative_folder_path(self):
        """

            Relative path to folder.

        Note:
            Ignoring the parent output folder.
        """
        return copy.deepcopy(self.__PROJECT.RELATIVE_PATH_TO_OUTPUT_FOLDER)



from eflow._hidden.parent_objects import DataPipelineSegment
from eflow.utils.pandas_utils import check_if_feature_exists
from eflow.utils.misc_utils import get_parameters

import copy

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

# Template
# def METHOD_NAME(self,
#                 df,
#                 df_features,
#                 '''ALL YOUR OTHER ARGS'''
#                 _add_to_que=True):

#
#     '''
#     YOUR CUSTOM CODE HERE
#     '''
#
#
#     # Add to the given pipeline segment
#     if _add_to_que:
#         params_dict = locals()
#          parameters = get_parameters(self.METHOD_NAME)
#          self._DataPipelineSegment__add_function_to_que("METHOD_NAME",
#                                                         parameters,
#                                                         params_dict)


class FeatureTransformer(DataPipelineSegment):
    """
        Combines, removes, scales, etc features of a pandas dataframe.
    """
    def __init__(self,
                 segment_id=None,
                 create_file=True):
        """
        Args:
            segment_id:
                Reference id to past segments of this object.

        Note/Caveats:
            When creating any public function that will be part of the pipeline's
            structure it is important to follow this given template. Also,
            try not to use _add_to_que. Can ruin the entire purpose of this
            project.
        """
        DataPipelineSegment.__init__(self,
                                     object_type=self.__class__.__name__,
                                     segment_id=segment_id,
                                     create_file=create_file)

    def remove_features(self,
                        df,
                        df_features,
                        feature_names,
                        _add_to_que=True):
        """

            Removes unwanted features from the dataframe and saves them to the
            pipeline segment structure if _add_to_que is set to True.

        Args:
            df:
                Pandas Dataframe to update.

            df_features:
                DataFrameTypes object to update.

            feature_names:
                Features to remove

            _add_to_que:
                Pushes the function to pipeline segment parent if set to 'True'.
        """

        if isinstance(feature_names, str):
            feature_names = [feature_names]

        for feature_n in feature_names:

            try:

                if feature_n in df_features.all_features():
                    df_features.remove_feature(feature_n)

                check_if_feature_exists(df,
                                        feature_n)
                df.drop(columns=[feature_n],
                        inplace=True)

            except KeyError:
                pass

        if _add_to_que:
            params_dict = locals()
            parameters = get_parameters(self.remove_features)
            self._DataPipelineSegment__add_function_to_que("remove_features",
                                                           parameters,
                                                           params_dict)


from eflow.utils.sys_utils import *
from eflow.utils.pandas_utils import df_to_image
from eflow.utils.image_processing_utils import create_plt_png
from eflow._hidden.parent_objects import ModelAnalysis
from eflow._hidden.custom_exceptions import RequiresPredictionMethods, ProbasNotPossible, \
    UnsatisfiedRequirments
from eflow.data_analysis import FeatureAnalysis

from eflow._hidden.constants import GRAPH_DEFAULTS

from sklearn.metrics import max_error
from sklearn.metrics import explained_variance_score
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_squared_log_error
from sklearn.metrics import median_absolute_error
from sklearn.metrics import r2_score

import scikitplot as skplt
import numpy as np
import warnings
import copy
import pandas as pd
from IPython.display import display
import matplotlib.pyplot as plt

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class RegressionAnalysis(ModelAnalysis):

    """
        Analyzes a classification model's result's based on the prediction
        function(s) passed to it. Creates graphs and tables to be saved in directory
        structure.
    """

    def __init__(self,
                 dataset_name,
                 model,
                 model_name,
                 feature_order,
                 target_feature,
                 pred_funcs_dict,
                 df_features,
                 project_sub_dir="Regression Analysis",
                 overwrite_full_path=None,
                 save_model=True,
                 notebook_mode=False):
        """
        Args:
            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            model:
                A fitted supervised machine learning model.

            model_name:
                The name of the model in string form.

            feature_order: collection object
                Features names in proper order to re-create the pandas dataframe.

            pred_funcs_dict:
                A dict of the name of the function and the function defintion for the
                model prediction methods.
                (Can handle either a return of probabilities or a singile value.)
                Init Example:
                pred_funcs = dict()
                pred_funcs["Predictions"] = model.predict
                pred_funcs["Probabilities"] = model.probas

            sample_data:
                Given data to then pass into our prediction functions to get a
                resultant to get the classification prediction 'type'.
                Can be a matrix or a vector.

            project_sub_dir:
                Creates a parent or "project" folder in which all sub-directories
                will be inner nested.

            overwrite_full_path:
                Overwrites the path to the parent folder.
            df_features:
                DataFrameTypes object; organizes feature types into groups.
        """

        # Init parent object
        ModelAnalysis.__init__(self,
                               f'{dataset_name}/{project_sub_dir}/Target Feature: {target_feature}/{model_name}',
                               overwrite_full_path)

        # Init objects without pass by refrence

        # Remove target feature from feature order when trying to recreate dataframe
        self.__target_feature = copy.deepcopy(target_feature)
        self.__feature_order = copy.deepcopy(feature_order)

        if self.__target_feature in self.__feature_order:
            self.__feature_order.remove(self.__target_feature)

        self.__model = copy.deepcopy(model)

        self.__model_name = copy.deepcopy(model_name)
        self.__pred_funcs_dict = copy.deepcopy(pred_funcs_dict)
        self.__df_features = copy.deepcopy(df_features)
        self.__notebook_mode = copy.deepcopy(notebook_mode)

        # Determines if the perform was called
        self.__called_from_perform = False

        # Attempt to save machine learning model
        try:
            if save_model:
                pickle_object_to_file(self.__model,
                                      self.folder_path,
                                      f'{self.__model_name}')
        except:
            pass
        # ---
        create_dir_structure(self.folder_path,
                             "_Extras")

        # Save features and or df_features object
        df_features.create_json_file_representation(self.folder_path + "_Extras",
                                                    "df_features.json")


    def get_predictions_names(self):
        return self.__pred_funcs_dict.keys()

    def perform_analysis(self,
                         X,
                         y,
                         dataset_name,
                         regression_error_analysis=False,
                         regression_correct_analysis=False,
                         ignore_metrics=[],
                         custom_metrics_dict=dict(),
                         display_visuals=True,
                         mse_score=None):
        """

            Runs all available analysis functions on the models predicted data.

        Args:
            X:
                Feature matrix.

            y:
                Target data vector.

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            regression_error_analysis: bool
                Perform feature analysis on data that was incorrectly predicted.

            regression_correct_analysis: bool
                Perform feature analysis on data that was correctly predicted.

            ignore_metrics:
                Specify the default metrics to not apply to the classification
                data_analysis.
                    *
                    *
                    *
                    *
                    *

            custom_metrics_dict:
                Pass the name of metric(s) with the function definition(s) in a
                dictionary.

            display_visuals:
                Controls visual display of error error data_analysis if it is able to run.

        Returns:
            Performs all classification functionality with the provided feature
            data and target data.
                * plot_precision_recall_curve
                * classification_evaluation
                * plot_confusion_matrix
        """
        try:
            self.__called_from_perform = True

            self.generate_matrix_meta_data(X,
                                           dataset_name + "/_Extras")

            print("\n\n" + "---" * 10 + f'{dataset_name}' + "---" * 10)



            for pred_name in self.__pred_funcs_dict.keys():

                self.regression_metrics(X,
                                        y,
                                        pred_name,
                                        dataset_name,
                                        display_visuals=display_visuals,
                                        ignore_metrics=ignore_metrics,
                                        custom_metrics_dict=custom_metrics_dict)

                if regression_error_analysis:
                    self.regression_error_analysis(X,
                                                   y,
                                                   pred_name,
                                                   dataset_name,
                                                   mse_score=mse_score,
                                                   display_print=False,
                                                   display_visuals=display_visuals)

                if regression_correct_analysis:
                    self.regression_correct_analysis(X,
                                                     y,
                                                     pred_name,
                                                     dataset_name,
                                                     mse_score=mse_score,
                                                     display_print=False,
                                                     display_visuals=display_visuals,)


        finally:
            self.__called_from_perform = False

    def regression_metrics(self,
                           X,
                           y,
                           pred_name,
                           dataset_name,
                           display_visuals=True,
                           save_file=True,
                           title="",
                           custom_metrics_dict=dict(),
                           ignore_metrics=[],
                           multioutput=[None,
                                        "uniform_average",
                                        "variance_weighted"]):
        """

            Creates a dataframe based on the prediction metrics
            of the feature matrix and target vector.

        Args:
            X:
                Feature matrix.

            y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned stored
                in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Display tables.

            save_file:
                Determines whether or not to save the generated document.

            title:
                Adds to the column 'Metric Score'.

            custom_metrics_dict:
                Pass the name of metric(s) and the function definition(s) in a
                dictionary.

            ignore_metrics:
                Specify the default metrics to not apply to the classification
                data_analysis.
                    * Precision
                    * MCC
                    * Recall
                    * F1-Score
                    * Accuracy

            average_scoring:
                Determines the type of averaging performed on the data.
                    * micro
                    * macro
                    * weighted

        Returns:
            Return a dataframe object of the metrics value.
        """
        filename = f'Metric Evaluation on {dataset_name} on {self.__model_name}'
        sub_dir = f'{dataset_name}/{pred_name}'

        if not isinstance(multioutput, list):
            multioutput = [multioutput]


        # Default metric name's and their function
        metric_functions = dict()
        metric_functions["Explained Variance Score"] = explained_variance_score
        metric_functions["Max Error"] = max_error
        metric_functions["Mean Absolute Error"] = mean_absolute_error
        metric_functions["Mean Squared Error"] = mean_squared_error
        metric_functions["Mean Squared Log Error"] = mean_squared_log_error
        metric_functions["Mean Squared Log Error"] = median_absolute_error
        metric_functions["R2 Score"] = r2_score
        warnings.filterwarnings('ignore')

        # Ignore default metrics if needed
        for remove_metric in ignore_metrics:
            if remove_metric in metric_functions:
                del metric_functions[remove_metric]

        # Add in custom metrics
        if len(custom_metrics_dict.keys()):
            metric_functions.update(custom_metrics_dict)

        # Evaluate model on metrics
        evaluation_report = dict()
        for metric_name in metric_functions:
            for multi in multioutput:

                model_predictions = self.__get_model_prediction(pred_name,
                                                                X)

                try:
                    if multi:
                        evaluation_report[f'{metric_name}({multi})'] = \
                            metric_functions[metric_name](y_true=y,
                                                          y_pred=model_predictions,
                                                          multioutput=multi)
                    else:
                        if metric_name not in evaluation_report.keys():
                            evaluation_report[f'{metric_name}'] = \
                                metric_functions[metric_name](y_true=y,
                                                              y_pred=model_predictions,
                                                              multioutput=multi)

                except TypeError:
                    if metric_name not in evaluation_report.keys():
                        evaluation_report[metric_name] = metric_functions[
                            metric_name](y,
                                         model_predictions)


        warnings.filterwarnings('default')

        if title and len(title) > 0:
            index_name = f"Metric Scores ({title})"
        else:
            index_name = "Metric Scores"

        # ---
        evaluation_report = pd.DataFrame({index_name:
                                              [f'{metric_score:.4f}'
                                               for metric_score
                                               in evaluation_report.values()]},
                                         index=list(evaluation_report.keys()))

        if display_visuals:
            if self.__notebook_mode:
                display(evaluation_report)
            else:
                print(evaluation_report)

        if save_file:
            df_to_image(evaluation_report,
                        self.folder_path,
                        sub_dir,
                        convert_to_filename(filename),
                        col_width=20,
                        show_index=True,
                        format_float_pos=4)

            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")


    def regression_correct_analysis(self,
                                    X,
                                    y,
                                    pred_name,
                                    dataset_name,
                                    mse_score,
                                    display_visuals=True,
                                    save_file=True,
                                    display_print=True,
                                    suppress_runtime_errors=True,
                                    aggregate_target_feature=True,
                                    selected_features=None,
                                    extra_tables=True,
                                    statistical_analysis_on_aggregates=True):
        """

            Compares the actual target value to the predicted value and performs
            analysis of all the data.

        Args:
            X: np.matrix or lists of lists
                Feature matrix.

            y: collection object
                Target data vector.

            pred_name: str
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name: str
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            feature_order: collection object
                Features names in proper order to re-create the pandas dataframe.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            extra_tables: bool
                When handling two types of features if set to true this will
                    generate any extra tables that might be helpful.
                    Note -
                        These graphics may create duplicates if you already applied
                        an aggregation in 'perform_analysis'

            aggregate_target_feature: bool
                Aggregate the data of the target feature if the data is
                non-continuous data.

                Note
                    In the future I will have this also working with continuous
                    data.

            selected_features: collection object of features
                Will only focus on these selected feature's and will ignore
                the other given features.

            statistical_analysis_on_aggregates: bool
                If set to true then the function 'statistical_analysis_on_aggregates'
                will run; which aggregates the data of the target feature either
                by discrete values or by binning/labeling continuous data.
        """
        model_predictions = self.__get_model_prediction(pred_name,
                                                        X)
        sub_dir = f'{dataset_name}/{pred_name}'

        if sum(model_predictions != y) == len(y):
            print("Your model predicted everything correctly for this dataset! No correct analysis needed!")
            print("Also sorry for your model...zero correct? Dam...")
        else:
            print("\n\n" + "*" * 10 +
                  "Generating graphs for when the model predicted correctly" +
                  "*" * 10 + "\n")

            all_mse_scores = []
            for i, pred in enumerate(model_predictions):
                all_mse_scores.append(mean_squared_error([pred], [y[i]]))

            # Generate error dataframe
            bool_list = np.array(all_mse_scores) < mse_score

            correct_df = pd.DataFrame.from_records(X[bool_list])
            correct_df.columns = self.__feature_order
            correct_df[self.__target_feature] = y[bool_list]

            # Directory path
            create_dir_structure(self.folder_path,
                                 sub_dir + f"/MSE score less than {mse_score}")
            output_path = f"{self.folder_path}/{sub_dir}/MSE score less than {mse_score}"

            # Create feature analysis
            feature_analysis = FeatureAnalysis(self.__df_features,
                                               overwrite_full_path=output_path)
            feature_analysis.perform_analysis(correct_df,
                                              dataset_name=dataset_name,
                                              target_features=[self.__target_feature],
                                              save_file=save_file,
                                              selected_features=selected_features,
                                              suppress_runtime_errors=suppress_runtime_errors,
                                              display_print=display_print,
                                              display_visuals=display_visuals,
                                              dataframe_snapshot=False,
                                              aggregate_target_feature=aggregate_target_feature,
                                              extra_tables=extra_tables,
                                              statistical_analysis_on_aggregates=statistical_analysis_on_aggregates)

    def regression_error_analysis(self,
                                  X,
                                  y,
                                  pred_name,
                                  dataset_name,
                                  mse_score,
                                  display_visuals=True,
                                  save_file=True,
                                  display_print=True,
                                  suppress_runtime_errors=True,
                                  aggregate_target_feature=True,
                                  selected_features=None,
                                  extra_tables=True,
                                  statistical_analysis_on_aggregates=True):
        """

            Compares the actual target value to the predicted value and performs
            analysis of all the data.

        Args:
            X: np.matrix or lists of lists
                Feature matrix.

            y: collection object
                Target data vector.

            pred_name: str
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name: str
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            feature_order: collection object
                Features names in proper order to re-create the pandas dataframe.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            extra_tables: bool
                When handling two types of features if set to true this will
                    generate any extra tables that might be helpful.
                    Note -
                        These graphics may create duplicates if you already applied
                        an aggregation in 'perform_analysis'

            aggregate_target_feature: bool
                Aggregate the data of the target feature if the data is
                non-continuous data.

                Note
                    In the future I will have this also working with continuous
                    data.

            selected_features: collection object of features
                Will only focus on these selected feature's and will ignore
                the other given features.

            statistical_analysis_on_aggregates: bool
                If set to true then the function 'statistical_analysis_on_aggregates'
                will run; which aggregates the data of the target feature either
                by discrete values or by binning/labeling continuous data.
        """

        # sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
        #                                                 dataset_name,
        #                                                 thresholds)

        model_predictions = self.__get_model_prediction(pred_name,
                                                        X)
        sub_dir = f'{dataset_name}/{pred_name}'

        if sum(model_predictions == y) == len(y):
            print("Your model predicted everything correctly for this dataset! No error analysis needed!")
        else:
            print("\n\n" + "*" * 10 +
                  "Generating graphs for when the model predicted incorrectly" +
                  "*" * 10 + "\n")

            all_mse_scores = []
            for i,pred in enumerate(model_predictions):
                all_mse_scores.append(mean_squared_error([pred],[y[i]]))

            # Generate error dataframe
            bool_list = np.array(all_mse_scores) > mse_score

            error_df = pd.DataFrame.from_records(X[bool_list])
            error_df.columns = self.__feature_order
            error_df[self.__target_feature] = y[bool_list]

            # Directory path
            create_dir_structure(self.folder_path,
                                 sub_dir + f"/MSE score greater than {mse_score}")
            output_path = f"{self.folder_path}/{sub_dir}/MSE score greater than {mse_score}"

            # Create feature analysis
            feature_analysis = FeatureAnalysis(self.__df_features,
                                               overwrite_full_path=output_path)
            feature_analysis.perform_analysis(error_df,
                                              dataset_name=dataset_name,
                                              target_features=[self.__target_feature],
                                              save_file=save_file,
                                              selected_features=selected_features,
                                              suppress_runtime_errors=suppress_runtime_errors,
                                              display_print=display_print,
                                              display_visuals=display_visuals,
                                              dataframe_snapshot=False,
                                              aggregate_target_feature=aggregate_target_feature,
                                              extra_tables=extra_tables,
                                              statistical_analysis_on_aggregates=statistical_analysis_on_aggregates)

    def __get_model_prediction(self,
                               pred_name,
                               X):

        if pred_name in self.__pred_funcs_dict.keys():
            return self.__pred_funcs_dict[pred_name](X)
        else:
            raise KeyError(f"No prediction name found of {pred_name}.")


from eflow._hidden.parent_objects import FileOutput
from eflow.utils.pandas_utils import df_to_image
from eflow.utils.image_processing_utils import create_plt_png
from eflow.utils.string_utils import convert_to_filename
from eflow.utils.sys_utils import create_dir_structure,correct_directory_path

import copy
from IPython.display import display
from matplotlib import pyplot as plt
import warnings
import numpy as np
import pandas as pd

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class ModelAnalysis(FileOutput):
    """
        All objects in model_analysis folder of eflow are related to this object.
    """

    def __init__(self,
                 project_name,
                 overwrite_full_path=None):
        """
        Args:
            project_name: string
                Sub directory to create on top of the directory
                'PARENT_OUTPUT_FOLDER_NAME'.

            overwrite_full_path: string
                The passed directory path must already exist. Will completely
                ignore the project name and attempt to point to this already
                created directory.
        """
        # Create/Setup project directory
        FileOutput.__init__(self,
                            project_name,
                            overwrite_full_path)

    def save_plot(self,
                  filename="Unknown filename",
                  sub_dir=""):
        """

            Checks the passed data to see if a plot can be saved; raises
            an error if it can't.

        Args:

            filename: string
                If set to 'None' will default to a pre-defined string;
                unless it is set to an actual filename.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

        """

        if not isinstance(sub_dir, str):
            raise TypeError(
                f"Expected param 'sub_dir' to be type string! Was found to be {type(sub_dir)}")

        # Create the png to save
        create_plt_png(self.folder_path,
                       sub_dir,
                       convert_to_filename(filename))


    def save_table_as_plot(self,
                           table,
                           filename="Unknown filename",
                           sub_dir="",
                           show_index=False):
        """

            Checks the passed data to see if a table can be saved as a plot;
            raises an error if it can't.

        Args:

            table: pd.Dataframe
                Dataframe object to convert to plot.

            filename: string
                If set to 'None' will default to a pre-defined string;
                unless it is set to an actual filename.

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.

            show_index: bool
                Show index of the saved dataframe.
        """

        if not isinstance(sub_dir, str):
            raise TypeError(
                f"Expected param 'sub_dir' to be type string! Was found to be {type(sub_dir)}")


        # Convert dataframe to plot
        df_to_image(table,
                    self.folder_path,
                    sub_dir,
                    convert_to_filename(filename),
                    show_index=show_index,
                    format_float_pos=2)

    def generate_matrix_meta_data(self,
                                  X,
                                  sub_dir):
        """

            Generates files/graphics in the proper directory for the matrix.

        Args:
            X: list of list; numpy array of numpy array or numpy matrix
                Numpy matrix

            sub_dir: string
                Specify the sub directory to append to the pre-defined folder path.
        """

        # Convert to numpy array if possible
        X = np.array(X)

        create_dir_structure(self.folder_path,
                             correct_directory_path(sub_dir + "/Meta Data"))

        output_folder_path = correct_directory_path(self.folder_path)

        # Create files relating to dataframe's shape
        shape_df = pd.DataFrame.from_dict({'Rows': [X.shape[0]],
                                           'Columns': [X.shape[1]]})
        df_to_image(shape_df,
                    f"{output_folder_path}/{sub_dir}",
                    "Meta Data",
                    "Matrix Shape Table",
                    show_index=False)


__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class DefaultException(Exception):
    def __init__(self,
                 error_message=None):
        self.__error_message = error_message

    def __str__(self):
        if not self.__error_message:
            return "eflow has raised an exception without describing it!"
        else:
            return self.__error_message

class SnapshotMismatchError(DefaultException):
    def __init__(self,
                 error_message=None):
        super().__init__(error_message=error_message)

class UnsatisfiedRequirments(DefaultException):
    def __init__(self,
                 error_message=None):
        super().__init__(error_message=error_message)

class PipelineError(DefaultException):
    def __init__(self,
                 error_message=None):
        super().__init__(error_message=error_message)

class PipelineSegmentError(DefaultException):
    def __init__(self,
                 error_message=None):
        super().__init__(error_message=error_message)


class RequiresPredictionMethods(DefaultException):
    def __init__(self,
                 error_message=None):
        super().__init__(error_message=error_message)


class ProbasNotPossible(DefaultException):
    def __init__(self,
                 error_message=None):
        super().__init__(error_message=error_message)



# class UnknownPredictionType(DefaultException):
#     def __init__(self):
#         super().__init__(error_message="Couldn't assert whether type was a"
#                                        " float,int,list,or numpy array.")
#
# class PredictionTypeUnknown(DefaultException):
#     def __init__(self):
#         super().__init__(error_message="Can not logical assert what data type"
#                                        " is being returned from the model.")
#
# class ProbabliltyPrediction(DefaultException):
#     def __init__(self):
#         super().__init__(error_message="Can not logically assert what data type"
#                                        " is being returned from the model!")
#
# class ProbasNotPossible(DefaultException):
#     def __init__(self):
#         super().__init__(error_message="Check if the model probas/pred"
#                                        " function you init can return"
#                                        " probabilities.")
#
# class ThresholdLength(DefaultException):
#     def __init__(self):
#         super().__init__(error_message="Thersholds list must have the same"
#                                        " length as the targets!")
#
# class ThresholdType(DefaultException):
#     def __init__(self):
#         super().__init__(error_message="Thersholds must be a list or"
#                                        " numpy array!")
#
# class RequiresPredictionMethods(DefaultException):
#     def __init__(self):
#         super().__init__(error_message="Prediction methods must be passed as a dictionary.")




from eflow._hidden.parent_objects import AutoModeler
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments
from eflow.utils.sys_utils import load_pickle_object,get_all_files_from_path, get_all_directories_from_path, pickle_object_to_file, create_dir_structure, write_object_text_to_file, json_file_to_dict, dict_to_json_file
from eflow.utils.eflow_utils import move_folder_to_eflow_garbage
from eflow.utils.modeling_utils import find_all_zscore_distances_from_target
from eflow.utils.math_utils import euclidean_distance
from eflow.data_analysis.feature_analysis import FeatureAnalysis
from eflow.data_pipeline_segments import DataEncoder

# Getting Sklearn Models
from sklearn.decomposition import PCA
from scipy.cluster.hierarchy import linkage, dendrogram,set_link_color_palette
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import AgglomerativeClustering
from sklearn.neighbors import kneighbors_graph

# Getting pyclustering
from pyclustering.cluster.kmeans import kmeans
from pyclustering.cluster.kmedians import kmedians
from pyclustering.cluster.kmedoids import kmedoids
from pyclustering.cluster.ema import ema
from pyclustering.cluster.cure import cure
from pyclustering.cluster.fcm import fcm
from pyclustering.cluster.somsc import somsc
from pyclustering.cluster.center_initializer import kmeans_plusplus_initializer,random_center_initializer

# Visuals libs
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import pylab as pl
import seaborn as sns
from IPython.display import display, HTML

# Misc
from collections import Counter
from scipy.stats import zscore
from kneed import KneeLocator
import pandas as pd
import six
import random
import numpy as np
import copy
import os
from tqdm import tqdm
import warnings


__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class AutoCluster(AutoModeler):
    """
        Analyzes the feature data of a pandas Dataframe object.
    """

    def __init__(self,
                 df,
                 project_sub_dir="",
                 project_name="Auto Clustering",
                 overwrite_full_path=None,
                 notebook_mode=False,
                 pca_perc=None):
        """
        Args:
            df: pd.Dataframe
                pd.Dataframe

            project_sub_dir: string
                Sub directory to write data.

            project_name: string
                Main project directory

            overwrite_full_path: string
                Overwrite full directory path to a given output folder

            notebook_mode: bool
                Display and show in notebook if set to true.
        """

        AutoModeler.__init__(self,
                             f'{project_sub_dir}/{project_name}',
                             overwrite_full_path)


        if os.path.exists(self.folder_path + "_Extras"):
            move_folder_to_eflow_garbage(self.folder_path+"_Extras",
                                         "Auto Clustering")


        # Define model
        self.__cluster_models_paths = dict()

        self.__notebook_mode = copy.deepcopy(notebook_mode)

        self.__models_suggested_clusters = dict()

        self.__pca = None
        self.__pca_perc = None

        self.__first_scaler = None
        self.__second_scaler = None
        self.__cutoff_index = None
        self.__pca_perc = pca_perc

        # --- Apply pca ---
        if pca_perc:

            # Create scaler object
            scaler = StandardScaler()
            scaled = scaler.fit_transform(df)

            self.__first_scaler = copy.deepcopy(scaler)

            print("\nInspecting scaled results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=df.columns,
                                          sub_dir="PCA",
                                          filename="Applied scaler results")

            pca, scaled = self.__visualize_pca_variance(scaled)

            self.__pca = pca

            # Generate "dummy" feature names
            pca_feature_names = ["PCA_Feature_" +
                                 str(i) for i in range(1,
                                                       len(df.columns) + 1)]

            print("\nInspecting applied scaler and pca results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=pca_feature_names,
                                          sub_dir="PCA",
                                          filename="Applied scaler and PCA results")

            if pca_perc < 1.0:
                # Find cut off point on cumulative sum
                cutoff_index = np.where(
                    pca.explained_variance_ratio_.cumsum() > pca_perc)[0][0]
            else:
                cutoff_index = scaled.shape[1] - 1

            print(
                "After applying pca with a cutoff percentage {0}%"
                " for the cumulative index. Using features 1 to {1}".format(
                    pca_perc, cutoff_index + 1))

            print("Old shape {0}".format(scaled.shape))

            scaled = scaled[:, :cutoff_index + 1]
            pca_feature_names = pca_feature_names[0: cutoff_index + 1]

            print("New shape {0}".format(scaled.shape))

            scaled = scaler.fit_transform(scaled)

            print("\nInspecting data after final scaler applied!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=pca_feature_names,
                                          sub_dir="PCA",
                                          filename="Applied final sclaer to process.")

            self.__second_scaler = copy.deepcopy(scaler)

            self.__scaled = scaled
            self.__cutoff_index = cutoff_index

        # Assumed PCA has already been applied; pass as matrix
        else:
            self.__scaled = df.values

        if os.path.exists(self.folder_path + "Models"):

            print("Found past models in directory structure! Attempting to re-initalize models...")

            cluster_directories = get_all_directories_from_path(self.folder_path + "Models")
            for cluster_dir in cluster_directories:
                if cluster_dir[0] != ".":
                    for dir in get_all_directories_from_path(self.folder_path + f"Models/{cluster_dir}"):
                        if dir[0] != ".":
                            for model_name in get_all_files_from_path(self.folder_path + f"Models/{cluster_dir}/{dir}",
                                                                      ".pkl"):

                                model_path = self.folder_path + f"Models/{cluster_dir}/{dir}/{model_name}"

                                try:
                                    obj = load_pickle_object(model_path)

                                    if isinstance(obj,list) or isinstance(obj, np.ndarray):
                                        continue
                                    else:
                                        del obj

                                    model_name = model_name.split(".")[0]

                                    self.__cluster_models_paths[
                                        model_name] = model_path
                                except Exception:
                                    pass

            if self.__cluster_models_paths:
                for model_name, model_path in self.__cluster_models_paths.items():
                    print(f"{model_name} was found at {model_path}\n")

        # Save objects to directory structure
        if self.__pca:
            pipeline_path = create_dir_structure(self.folder_path,
                                                 "Data Pipeline")


            # Pickle data pipeline objects
            pickle_object_to_file(self.__pca,
                                  pipeline_path,
                                  "PCA")

            pickle_object_to_file(self.__first_scaler,
                                  pipeline_path,
                                  "First Scaler")

            pickle_object_to_file(self.__second_scaler,
                                  pipeline_path,
                                  "First Scaler")

            pickle_object_to_file(self.__pca_perc,
                                  pipeline_path,
                                  "PCA Percentage")

            # Save Dimensions and Cutoff Index
            write_object_text_to_file(self.__cutoff_index,
                                      pipeline_path,
                                      "Cutoff Index")

            write_object_text_to_file(self.__cutoff_index + 1,
                                      pipeline_path,
                                      "Dimensions")



    # --- Getters/Setters
    def get_scaled_data(self):
        """

            Get a copy of the stored data

        Returns:
            Returns the stored data
        """
        return copy.deepcopy(self.__scaled)

    def get_all_cluster_models(self):
        """

            Gets the model names and model instances in dictionary form.

        Return:
            Returns the model name to model instance dict
        """
        tmp_dict = dict()
        for model_name, model_path in self.__cluster_models_paths.items():

            try:

                model = load_pickle_object(model_path)

                tmp_dict[model_name] = model
            except EOFError:
                pass

        return tmp_dict

    def delete_scaled_data(self):
        """
            Removes the matrix data in order to save RAM when running
            analysis on the actual data.
        """
        del self.__scaled
        self.__scaled = None

    def apply_clustering_data_pipeline(self,
                                       data):
        """
            Apply the scaler, dimension reduction transformation, matrix shrink
            and second scaler to the data.

        Args:
            data: np.matrix, list of lists, pd.DataFrame
                Data that is similar in form and value structure to the data passed
                on initialization.

        Returns:
            Returns back data after transformations are applied.
        """

        data = self.__first_scaler.transform(data)
        data = self.__pca.transform(data)
        data = data[:, :self.__cutoff_index + 1]
        data = self.__second_scaler.transform(data)

        return data

    def get_model(self,
                  model_name):
        """
            Get the model object from the stored dirctories based on the model
            name.

        Args:
            model_name: str
                Name of the model that references the saved model in
                the provided directory.

        Returns:
            Gives back the model object.
        """

        return load_pickle_object(self.__cluster_models_paths[model_name])


    def visualize_hierarchical_clustering(self,
                                          linkage_methods=None,
                                          display_print=True,
                                          display_visuals=True):
        """
            Displays hierarchical cluster graphs with provided methods.

        Args:
            linkage_methods: list of strings
                All methods applied to the linkage

            display_print: bool
                Display print outputs

            display_visuals: bool
                Display plot data if set to ture.

        Returns:
            All cluster counts found.
        """

        best_clusters = []

        # ---
        if not linkage_methods:
            linkage_methods = ["complete",
                               "single",
                               "weighted",
                               "ward",
                               "average",
                               "centroid",
                               "median"]

        # Apply methods to each dendrogram
        for method in linkage_methods:

            if display_print:
                print(f"Creating graphic for Hierarchical Clustering Method: {method}...")

            # Create mergings
            mergings = linkage(self.__scaled, method=method)

             # {"Yellow":"#d3d255",
             # "Magenta":"#c82bc9",
             # "Black":"#030303",
             # "Red":"#ff403e",
             # "Green":"#3f9f3f",
             # "Cyan":"#0ec1c2",
             # "Brown": "#775549",
             # "Silver": "#C0C0C0",
             # "Blue":"#24326f",
             # "Orange":"#cc7722",
             # "Mauve":"#9c7c8c"}

            # Set plot
            plt.figure(figsize=(12, 7))
            set_link_color_palette(None)

            # Plot the dendrogram, using varieties as labels
            color_list = dendrogram(mergings,
                                    labels=list(range(0, len(self.__scaled,))),
                                    leaf_rotation=90,
                                    leaf_font_size=3)["color_list"]

            method = method.capitalize()
            plt.title(f"Hierarchical Clustering Method : \'{method}\'")

            # -----
            self.save_plot("Hierarchical Clustering",
                           f"Hierarchical Clustering Method {method} without legend")

            del mergings


            # Create proper cluster names based on the color names
            color_cluster_count = dict()
            last_color = None
            known_colors = set()
            color_cluster_order = list()
            seq_len = 0
            i = 0

            # -----
            for color in copy.deepcopy(color_list):

                # Proper color name
                color = self.__get_color_name(color)

                # Name for old cluster color sequence found
                if color in known_colors:
                    color_list[i] = f"{color} cluster {color_cluster_count[color]}"

                # Name for new cluster color sequence found
                else:
                    color_list[i] = f"{color} cluster 0"

                # Track unique cluster color order
                if color_list[i] not in color_cluster_order:
                    color_cluster_order.append(color_list[i])

                if last_color:
                    # Sequence of color has yet to be broken
                    if last_color == color:

                        # Only really need to check if the sequence has a length of 1
                        if seq_len <= 2:
                            seq_len += 1

                    # Sequence break
                    else:

                        # Valid cluster found
                        if seq_len > 1:

                            # Track all known color names
                            if last_color not in known_colors:
                                known_colors.add(last_color)

                            # If color is repeated; then make a new cluster count name
                            if last_color not in color_cluster_count.keys():
                                color_cluster_count[last_color] = 1
                            else:
                                color_cluster_count[last_color] += 1

                        # Invalid color cluster found; apply needed changes
                        else:
                            color_list.pop(i-1)
                            i -= 1

                        seq_len = 0

                last_color = color
                i += 1

            # Create legend for each cluster name and the amount of per sample
            counter_object = Counter(color_list)
            cluster_color_count = dict()
            handles = []

            # Make sure the legend is in the same order the cluster appear in the dendrogram
            for color_cluster_name in color_cluster_order:
                if color_cluster_name in counter_object.keys():
                    cluster_color_count[color_cluster_name] = counter_object[color_cluster_name]
                    try:
                        handles.append(mpatches.Patch(
                            color=color_cluster_name.split(" cluster ")[0],
                            label=color_cluster_name + f": {counter_object[color_cluster_name]} samples"))

                    except:
                        handles.append(mpatches.Patch(
                            color="black",
                            label=color_cluster_name + f": {counter_object[color_cluster_name]} samples"))

            # Plot the legend and save the plot
            plt.legend(handles=handles,
                       loc='upper right',
                       bbox_to_anchor=(1.32, 1.01),
                       title=f"Clusters ({len(handles)})")

            best_clusters.append(len(handles))
            if display_visuals and self.__notebook_mode:
                plt.show()

            self.save_plot("Hierarchical Clustering",
                           f"Hierarchical Clustering Method {method} with legend")

            plt.close('all')

        # Save results into _Extras folder
        best_clusters.sort()
        self.__models_suggested_clusters["Hierarchical Clustering"] = best_clusters
        self.__save_update_best_model_clusters()
        return best_clusters


    # def __visualize_clusters(self, model, output_path, model_name=""):
    #     """
    #         Creates visualization of clustering model on given data.
    #     """
    #     markers = ["+", "*", "X", "o", "v", "P", "H", "4", "p", "D", "s",
    #                "1", "x", "d", "_"]
    #     colors = ['b', 'g', 'r', 'c', 'm', 'y',
    #               '#007BA7', '#ff69b4', '#CD5C5C', '#7eab19', '#1a4572',
    #               '#2F4F4F', '#4B0082', '#d11141', '#5b2504']
    #
    #     # Display ranking on color based on amount data points per cluster
    #     unique, counts = np.unique(model.labels_, return_counts=True)
    #     cluster_names = ["Cluster:" + str(cluster_label)
    #                      for cluster_label in unique]
    #     self.__display_rank_graph(feature_names=cluster_names,
    #                               metric=counts,
    #                               title=model_name,
    #                               output_path=output_path,
    #                               model_name=model_name,
    #                               y_title="Clusters",
    #                               x_title="Found per cluster")
    #     pl.figure(figsize=(8, 7))
    #
    #     # Display clustered graph
    #     cluster_array = list(range(0, len(cluster_names)))
    #     scaled_cluster_label = np.hstack(
    #         (self.__scaled, np.reshape(
    #             model.labels_.astype(int), (self.__scaled.shape[0], 1))))
    #     for i in range(0, scaled_cluster_label.shape[0]):
    #         cluster_label = int(scaled_cluster_label[i][-1])
    #         cluster_array[cluster_label] = pl.scatter(
    #             scaled_cluster_label[i, 0], scaled_cluster_label[i, 1],
    #             c=colors[cluster_label], marker=str(markers[cluster_label]))
    #
    #     pl.legend(cluster_array, cluster_names)
    #     pl.title(model_name + ' visualized with data', fontsize=15)
    #     self.__image_processing_utils(output_path,
    #                           model_name + "_Visualized_Cluster")
    #     plt.show()
    #     plt.close()
    #     pl.close()
    def create_elbow_models(self,
                            model_names=["K-Means",
                                         "K-Medians",
                                         "K-Medoids",
                                         "Somsc",
                                         "Cure",
                                         "Fuzzy C-Means"],
                            sequences=3,
                            max_k_value=15,
                            display_visuals=True):
        """
            Create multiple sequences of defined clusters and their related
            inertia values to find the most agreed

        Args:
            model_names: list of strings
                Defines which models should be created.

            sequences: int
                How many model sequences to create

            max_k_value: int
                Max clusters to create

            display_visuals: bool
                Display graphics and tables if set to True.
        """

        model_names = set(model_names)

        # Model names and their model instances
        names_model_dict = {"K-Means":kmeans,
                            "K-Medians":kmedians,
                            "K-Medoids":kmedoids,
                            "Somsc":somsc,
                            "Cure":cure,
                            "Fuzzy C-Means": fcm}

        # Iterate through passed model names
        for name in model_names:

            if name in names_model_dict.keys():

                # Only requires 1 elbow sequence needed
                if name == "Somsc" or name == "Cure":
                    best_clusters = self.__create_elbow_seq(name,
                                                            names_model_dict[name],
                                                            sequences=1,
                                                            max_k_value=max_k_value,
                                                            display_visuals=display_visuals)
                else:
                    best_clusters = self.__create_elbow_seq(name,
                                                            names_model_dict[name],
                                                            sequences=sequences,
                                                            max_k_value=max_k_value,
                                                            display_visuals=display_visuals)

                # Save cluster results in
                best_clusters.sort()
                self.__models_suggested_clusters[name] = best_clusters
                self.__save_update_best_model_clusters()
            else:
                raise UnsatisfiedRequirments(f"Unknown model name passed: \"{name}\"")

        return best_clusters

    def create_agglomerative_models(self,
                                    n_cluster_list,
                                    linkage_methods=None):
        """

            Create multiple agglomerative models based on a list of
            'n_clusters' values and defined linkage methods.
        """
        raise ValueError("This functionality has been clamped!...Hopefully someone get's this refrence")
        pass
        # if isinstance(n_cluster_list, int):
        #     n_cluster_list = [n_cluster_list]
        #
        # if not linkage_methods:
        #     linkage_methods = ["ward", "complete", "average", "single"]
        #
        # knn_graph = kneighbors_graph(
        #     self.__scaled, len(
        #         self.__scaled) - 1, include_self=False)
        #
        # for n_clusters in n_cluster_list:
        #     for connectivity in (None, knn_graph):
        #
        #         for _, linkage in enumerate(linkage_methods):
        #             model = AgglomerativeClustering(linkage=linkage,
        #                                             connectivity=connectivity,
        #                                             n_clusters=n_clusters)
        #             model.fit(self.__scaled)
        #             self.__cluster_models[
        #                 "AgglomerativeClustering_{0}_"
        #                 "cluster{1}_Connectivity{2}".format(
        #                     linkage,
        #                     n_clusters, connectivity is not None)] = model
        #
        #             print(
        #                 "Successfully generate Agglomerative model with "
        #                 "linkage {0} on n_clusters={1}".format(
        #                     linkage, n_clusters))

        # self.__models_suggested_clusters["Agglomerative models"] =

    def evaluate_all_models(self,
                            df,
                            df_features,
                            qualitative_features=[],
                            zscore=None):
        """

            Loop through all models and evaluate the given model with
            'evaluate_model'. Read 'evaluate_model' to learn more.

        Args:
            df: pd.DataFrame
                Pandas dataframe

            df_features: DataFrameTypes object from eflow.
                DataFrameTypes object; organizes feature types into groups.

            zscore: float, int, None
                Will find the distances between given the given clusters then
                we apply data pipeline.
        """

        for model_name, model_path in self.__cluster_models_paths.items():
            self.evaluate_model(model_name=model_name,
                                df=df,
                                df_features=df_features,
                                qualitative_features=qualitative_features,
                                zscore=zscore)
            print("------" * 10)

    def evaluate_model(self,
                       model_name,
                       df,
                       df_features,
                       qualitative_features=[],
                       zscore=None,
                       target_features=None,
                       display_visuals=False,
                       display_print=False,
                       suppress_runtime_errors=True,
                       aggregate_target_feature=True,
                       selected_features=None,
                       extra_tables=True,
                       statistical_analysis_on_aggregates=True):
        """

            The main purpose of 'evaluate_model' is to display/save tables/plots
            accoiated with describing the model's 'findings' for each cluster.

        Args:

            model_name: str
                The string key to give the dict

            df: pd.Dataframe
                Dataframe object

            df_features: DataFrameTypes object from eflow.
                DataFrameTypes object; organizes feature types into groups.

            qualitative_features: list
                Any features that are currently dummy encoded will be reversed
                when doing feature analysis.

            zscore: float, int, None, or list of ints/floats
                After calculating each distance to the center point; a zscore
                value for distance will be found. Setting this parameter defines
                a cut off point where data exceeding a certain zscore will be
                ignored when doing FeatureAnalysis

            target_features: collection of strings or None
                A feature name that both exists in the init df_features
                and the passed dataframe.

                Note
                    If init to 'None' then df_features will try to extract out
                    the target feature.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            extra_tables: bool
                When handling two types of features if set to true this will
                    generate any extra tables that might be helpful.
                    Note -
                        These graphics may create duplicates if you already applied
                        an aggregation in 'perform_analysis'

            aggregate_target_feature: bool
                Aggregate the data of the target feature if the data is
                non-continuous data.

                Note
                    In the future I will have this also working with continuous
                    data.

            selected_features: collection object of features
                Will only focus on these selected feature's and will ignore
                the other given features.

            statistical_analysis_on_aggregates: bool
                If set to true then the function 'statistical_analysis_on_aggregates'
                will run; which aggregates the data of the target feature either
                by discrete values or by binning/labeling continuous data.
        """

        # Get/Set model related data
        model = load_pickle_object(self.__cluster_models_paths[model_name])
        model_path = self.__cluster_models_paths[model_name]
        model_dir = os.path.dirname(model_path)
        model_sub_dir = model_dir.replace(self.folder_path,"",1)
        center_points = load_pickle_object(os.path.dirname(
            self.__cluster_models_paths[model_name]) + "/Center points.pkl")
        cluster_labels = json_file_to_dict(f"{model_dir}/Cluster Labels.json")

        # Store single instance of int/float/None in a list
        if isinstance(zscore,float) or isinstance(zscore,type(None)) or isinstance(zscore,int):
            zscore = [zscore]

        # Iterate through all zscore values
        for z_val in zscore:

            cluster_labels_lengths = dict()

            # Handling all zscore values
            no_zscore = False

            if not z_val:
                z_val = float("inf")
                no_zscore = True
            else:
                z_val = round(z_val, 5)

            print(f"Model Name: {model_name}")
            print(f"Clusters: {len(model.get_clusters())}")
            print(f"Distance Zscore: {z_val}")
            print()

            for i,cluster_indexes in enumerate(model.get_clusters()):

                label = cluster_labels[str(i)]

                print(f"Feature Analysis on cluster {label}")

                # Create cluster dataframe
                tmp_df = df.loc[cluster_indexes].reset_index(drop=True)

                # Find and apply zscore cut off point
                bool_array = find_all_zscore_distances_from_target(self.apply_clustering_data_pipeline(tmp_df.values),
                                                                   center_points[i])

                bool_array = (bool_array >= -z_val) & (bool_array <= z_val)

                removed_dp_count = len(bool_array) - bool_array.sum()
                removed_dp_percentage = removed_dp_count / len(tmp_df)

                tmp_df = tmp_df[bool_array].reset_index(drop=True)

                del bool_array

                # Decode and revert dummies data
                tmp_df_features = copy.deepcopy(df_features)

                data_encoder = DataEncoder(create_file=False)

                data_encoder.revert_dummies(tmp_df,
                                            tmp_df_features,
                                            qualitative_features=qualitative_features)

                data_encoder.decode_data(tmp_df,
                                         tmp_df_features,
                                         apply_value_representation=False)

                data_encoder.apply_value_representation(tmp_df,
                                                        df_features)
                del data_encoder

                cluster_labels_lengths[label] = len(tmp_df)

                # Handle naming/managing the directory structure
                if no_zscore:
                    create_dir_structure(directory_path=model_dir,
                                         create_sub_dir="No Distance Zscore")

                    self.__create_cluster_profile(tmp_df,
                                                  tmp_df_features,
                                                  sub_dir=f"{model_sub_dir}/No Distance Zscore/Cluster: {label}")


                    feature_analysis = FeatureAnalysis(tmp_df_features,
                                                       overwrite_full_path=f"{model_dir}/No Distance Zscore",
                                                       notebook_mode=self.__notebook_mode)
                else:
                    create_dir_structure(directory_path=model_dir,
                                         create_sub_dir=f"Distance Zscore = {z_val}/Cluster: {label}")

                    self.__create_cluster_profile(tmp_df,
                                                  tmp_df_features,
                                                  sub_dir=f"{model_sub_dir}/Distance Zscore = {z_val}/Cluster: {label}")


                    feature_analysis = FeatureAnalysis(tmp_df_features,
                                                       overwrite_full_path=f"{model_dir}/Distance Zscore = {z_val}",
                                                       notebook_mode=self.__notebook_mode)

                # Analyze the cluster dataframe
                feature_analysis.perform_analysis(tmp_df,
                                                  dataset_name=f"Cluster: {label}",
                                                  target_features=target_features,
                                                  display_visuals=display_visuals,
                                                  display_print=display_print,
                                                  dataframe_snapshot=True,
                                                  save_file=True,
                                                  suppress_runtime_errors=suppress_runtime_errors,
                                                  aggregate_target_feature=aggregate_target_feature,
                                                  selected_features=selected_features,
                                                  extra_tables=extra_tables,
                                                  statistical_analysis_on_aggregates=statistical_analysis_on_aggregates)

                # Save removed data points
                write_object_text_to_file(removed_dp_count,
                                          f"{feature_analysis.folder_path}/Cluster: {label}",
                                          filename="Removed dp count")

                write_object_text_to_file(removed_dp_percentage,
                                          f"{feature_analysis.folder_path}/Cluster: {label}",
                                          filename="Removed dp percentage")

            if no_zscore:
                self.__display_cluster_label_rank_graph(list(cluster_labels_lengths.keys()),
                                                        list(cluster_labels_lengths.values()),
                                                        sub_dir=f"{model_dir}/No Distance Zscore",
                                                        display_visuals=self.__notebook_mode)
            else:
                self.__display_cluster_label_rank_graph(list(cluster_labels_lengths.keys()),
                                                        list(cluster_labels_lengths.values()),
                                                        sub_dir=f"{model_sub_dir}/Distance Zscore = {z_val}",
                                                        display_visuals=self.__notebook_mode)

            print("###" * 4)


    def __create_cluster_profile(self,
                                 df,
                                 df_features,
                                 sub_dir):
        """

            Generate a cluster profile based on the clustered data. A cluster
            profile gets the mean of a numerical series data and the mode of
            a non-numerical one.

        Args:
            df: pd.DataFrame
                Dataframe object

            df_features: DataFrameTypes object from eflow.
                DataFrameTypes object; organizes feature types into groups.

            sub_dir: string
                Sub directory to create when writing data.
        """
        cluster_profile_dict = dict()

        for feature_name in df.columns:

            if feature_name not in df_features.all_features():
                raise ValueError(f"No feature named {feature_name} was found in df_features.")

            if feature_name in df_features.continuous_numerical_features():
                cluster_profile_dict[feature_name] = df[feature_name].mean()
            else:
                cluster_profile_dict[feature_name] = ":".join([str(i) for i in df[feature_name].mode()])
                if feature_name in df_features.bool_features():
                    if cluster_profile_dict[feature_name] == "1":
                        cluster_profile_dict[feature_name] = True
                    elif cluster_profile_dict[feature_name] == "0":
                        cluster_profile_dict[feature_name] = False

        self.save_table_as_plot(pd.DataFrame(cluster_profile_dict, index=[0]).transpose(),
                                sub_dir=sub_dir,
                                filename="Cluster Profile",
                                show_index=True)

    def __inspect_feature_matrix(self,
                                 matrix,
                                 feature_names,
                                 sub_dir,
                                 filename):
        """

            Creates a dataframe to quickly analyze a matrix of data by mean and
            standard deviation.

        Args:

            matrix: list of lists, np.matrix, pd.DataFrame
                Matrix data to convert to mean and

            feature_names: list of strings
                Each dimension's name.

            sub_dir: string
                Sub directory to create when writing data.

            filename: string
                Name of the file
        """

        # Calculate mean and std for each dimension
        mean_matrix = np.mean(matrix, axis=0)
        std_matrix = np.std(matrix, axis=0)

        # Create relationship dict
        data_dict = dict()
        for index, feature_name in enumerate(feature_names):
            data_dict[feature_name] = [mean_matrix[index],
                                       std_matrix[index]]

        # Convert to dataframe, display, and save to directory.
        tmp_df = pd.DataFrame.from_dict(data_dict,
                                        orient='index',
                                        columns=['Mean', 'Standard Dev'])

        if self.__notebook_mode:
            display(tmp_df)
        else:
            print(tmp_df)

        self.save_table_as_plot(tmp_df,
                                sub_dir=sub_dir,
                                filename=filename,
                                show_index=True,
                                format_float_pos=5)


    def __visualize_pca_variance(self,
                                 data):
        """

            Visualize PCA matrix feature importance.

        Args:
            data: list of list, np.matrix,
                Values to have pca applied too.

        Credit to favorite teacher Narine Hall for making this function.
        I wouldn't be the programmer I am today if it wasn't for her.
        """

        # Check for pca variance
        pca = PCA()
        data = pca.fit_transform(data)

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_)
        plt.xticks()
        plt.ylabel('Variance ratio')
        plt.xlabel('PCA feature')
        plt.tight_layout()

        self.save_plot("PCA",
                       "PCA Feature Variance Ratio")

        if self.__notebook_mode:
            plt.show()
            plt.close("all")

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_.cumsum())
        plt.xticks()
        plt.ylabel('Cumulative sum of variances')
        plt.xlabel('PCA feature')
        plt.tight_layout()
        self.save_plot("PCA",
                       "PCA Cumulative Sum of Variances")

        if self.__notebook_mode:
            plt.show()
            plt.close("all")

        return pca, data


    def __create_elbow_seq(self,
                           model_name,
                           model_instance,
                           sequences,
                           max_k_value,
                           display_visuals):
        """

            Fit's multiple cluster models and calculates each model's intertia
            value. Because the clustering is subjective; finding the 'elbow' of
            the elbow determines the 'best' model. That job is passed on to
            "__find_best_elbow_models" to find the elbow of each sequence and
            then the best matching elbow of all elbows.

        Args:
            model_name: string
                Model's name

            model_instance: pyclustering model
                Model's instance

            sequences: int
                How many model sequences to create

            max_k_value: int
                How long to create the model sequence

            display_visuals: bool
                Display the graphics

        Returns:
            Get's the best cluster amounts
        """

        max_k_value += 1

        # Matrix declarations
        k_models = []
        inertias = []

        for elbow_seq_count in range(0,sequences):
            tmp_inertias = []
            tmp_k_models = []

            # Set up progress bar
            if display_visuals:
                pbar = tqdm(range(1,max_k_value), desc=f"{model_name} Elbow Seq Count {elbow_seq_count + 1}")
            else:
                pbar = range(1,max_k_value)

            # Find random center points for clusters
            initial_centers = self.__create_random_initial_centers(model_name,
                                                                   max_k_value)
            for k_val in pbar:

                if display_visuals:
                    pbar.set_postfix(model_count=k_val,
                                     refresh=True)

                model = self.__create_pyclustering_model(model_name=model_name,
                                                         model_instance=model_instance,
                                                         initial_centers=initial_centers,
                                                         k_val=k_val)

                # Run cluster analysis and obtain results.
                model.process()
                final_centers = np.array(self.__get_centers(model))
                labels = [self.__nearest_cluster(final_centers, dp)
                          for dp in self.__scaled]

                inertia = sum(((final_centers[l] - x) ** 2).sum()
                              for x, l in zip(self.__scaled, labels))

                # Append the inertia to the list of inertias
                tmp_inertias.append(inertia)
                tmp_k_models.append(model)

            k_models.append(tmp_k_models)
            inertias.append(tmp_inertias)

        return self.__find_best_elbow_models(model_name,
                                             k_models,
                                             inertias,
                                             display_visuals)

    def __get_centers(self,
                      model):
        """

            Get's/creates the center point of each cluster.
        Args:
            model: pyclustering
                Clustering model instance.

        Returns:
            The center points of the clusters
        """
        try:
            return model.get_centers()
        except AttributeError:

            center_points = []

            for cluster_indexes in model.get_clusters():
                all_dps = np.matrix([self.__scaled[i] for i in cluster_indexes])
                center_dp = all_dps.mean(0)

                # Grave Yard code: Use existing point rather than generating abstract average data point
                # np.absolute(all_dps - center_dp).sum(1).argmin()

                center_points.append(np.array(center_dp.tolist()[0]))

            return center_points


    def __create_pyclustering_model(self,
                                    model_name,
                                    model_instance,
                                    initial_centers,
                                    k_val):
        """

            Generates simple clustering model's that only require the clustering
            amount.

        Args:
            model_name: string
                Name of the model

            model_instance: pyclustering
                Clustering model

            initial_centers: list of list of floats
                Center points of each cluster

            k_val:
                Value of how many cluster's that should be generated.

        Returns:
            Returns a init instance of the needed pyclustering model.
        """


        if model_name == "Somsc" or model_name == "Cure":
            model = model_instance(self.__scaled,
                                   k_val)
        else:
            model = model_instance(self.__scaled,
                                   initial_centers[0:k_val])
        return model

    def __create_random_initial_centers(self,
                                        model_name,
                                        k_val):
        """

            Generates multiple random starting center points.

        Args:
            model_name: string
                Name of the given model (some models have different requirements)

            k_val: int
                How many random center points/clusters need to be created.
        """
        if model_name == "Somsc" or model_name == "Cure":
            return None
        elif model_name == "K-Medoids":
            return [i for i in self.__get_unique_random_indexes(k_val)]

        elif model_name == "K-Means" or model_name == "Fuzzy C-means":
            return kmeans_plusplus_initializer(self.__scaled,
                                               k_val).initialize()
        else:
            return random_center_initializer(self.__scaled,
                                             k_val).initialize()

    def __find_best_elbow_models(self,
                                 model_name,
                                 k_models,
                                 inertias,
                                 display_visuals=True):
        """

            Find the elbow of each sequence and then find the best matching
            elbow that all sequences agree upon.

        Args:
            model_name: string
                Name of all model instance

            k_models: list of list of pyclustering models
                All pyclustering models

            inertias: list of list of floats
                All model's intertia's

            display_visuals: bool
                Determines if graphics should be displayed

        Returns:
            Returns back the best numbers of clusters.
        """

        ks = range(1,
                   len(inertias[0]) + 1)

        # Set graphic info
        plt.figure(figsize=(13, 6))
        plt.title(f"All possible {model_name} Elbow's", fontsize=15)
        plt.xlabel('Number of clusters, k')
        plt.ylabel('Inertia')
        plt.xticks(ks)

        # ----
        elbow_inertias_matrix = None
        inertias_matrix = None
        elbow_models = []
        elbow_sections = []
        center_elbow_count = dict()
        proximity_elbow_count = dict()
        elbow_cluster = None

        # Plot inertias against ks (y values against x values)
        for i in range(0,len(inertias)):

            # Determine curve
            curve = self.__determine_curve(inertias[i])

            if curve == "concave":
                online = True
            else:
                online = False

            # Find the best cluster model
            elbow_cluster = KneeLocator(ks,
                                        inertias[i],
                                        curve="convex",
                                        online=online,
                                        direction='decreasing').elbow
            # print(f"elbow_cluster:{elbow_cluster}")

            if elbow_cluster == 1 or not elbow_cluster:
                print("Elbow was either one or None for the elbow seq.")
                continue

            # Plot sequence
            plt.plot(ks,
                     inertias[i],
                     '-o',
                     color='#367588',
                     alpha=0.5)

            # Count all center elbows found
            if str(elbow_cluster) not in center_elbow_count.keys():
                center_elbow_count[str(elbow_cluster)] = 1
            else:
                center_elbow_count[str(elbow_cluster)] += 1

            # Get elbow section (1 before center and 1 after center)
            for k_val in [elbow_cluster - 1, elbow_cluster, elbow_cluster + 1]:
                elbow_sections.append([ks[k_val - 1],inertias[i][k_val - 1]])

                if str(k_val) not in proximity_elbow_count.keys():
                    proximity_elbow_count[str(k_val)] = 1
                else:
                    proximity_elbow_count[str(k_val)] += 1

            # Get inertia values
            if isinstance(elbow_inertias_matrix, type(None)):
                inertias_matrix = np.matrix([inertias[i]])
                elbow_inertias_matrix = np.matrix(inertias[i][elbow_cluster - 2:elbow_cluster + 1])
            else:
                inertias_matrix = np.vstack([inertias_matrix, inertias[i]])

                elbow_inertias_matrix = np.vstack(
                    [elbow_inertias_matrix, inertias[i][elbow_cluster - 2:elbow_cluster + 1]])

            elbow_models.append(k_models[i][elbow_cluster - 2:elbow_cluster + 1])

        # Plot elbow sections
        for elbow_s in elbow_sections:
            k_val = elbow_s[0]
            intertia = elbow_s[1]
            plt.plot(k_val,
                     intertia,
                     'r*',)

        del inertias
        del k_models
        del elbow_cluster

        self.save_plot(f"Models/{model_name}",
                       f"All possible {model_name} Elbow's",)

        if display_visuals and self.__notebook_mode:
            plt.show()

        plt.close("all")

        # Get the counts of the center elbows
        center_elbow_count = pd.DataFrame({"Main Elbows": list(center_elbow_count.keys()),
                                           "Counts": list(center_elbow_count.values())})
        center_elbow_count.sort_values(by=['Counts'],
                                       ascending=False,
                                       inplace=True)

        self.save_table_as_plot(
            center_elbow_count,
            sub_dir=f"Models/{model_name}",
            filename="Center Elbow Count")

        # Get the counts of the proximity elbows
        proximity_elbow_count = pd.DataFrame({"Proximity Elbow": list(proximity_elbow_count.keys()),
                                              "Counts": list(proximity_elbow_count.values())})
        proximity_elbow_count.sort_values(by=['Counts'],
                                          ascending=False,
                                          inplace=True)

        self.save_table_as_plot(
            proximity_elbow_count,
            sub_dir=f"Models/{model_name}",
            filename="Proximity Elbow Count")

        # Set up graphic
        plt.figure(figsize=(13, 6))
        plt.title(f"Best of all {model_name} Elbows", fontsize=15)
        plt.xlabel('Number of clusters, k')
        plt.ylabel('Inertia')
        plt.xticks(ks)



        # Find the most agreed upon elbow section
        average_elbow_inertias = elbow_inertias_matrix.mean(0)

        elbow_vote = []
        for vector in elbow_inertias_matrix:
            elbow_vote.append(
                np.absolute(vector - average_elbow_inertias).sum())

        best_elbow_index = np.array(elbow_vote).argmin()

        print(inertias_matrix[best_elbow_index].tolist()[0])

        # Plot most agreed upon elbow section
        plt.plot(ks,
                 inertias_matrix[best_elbow_index].tolist()[0],
                 '-o',
                 color='#367588')

        # Save the best cluster models
        best_clusters = []
        for model in elbow_models[best_elbow_index]:
            k_val = len(model.get_clusters())

            model_path = create_dir_structure(self.folder_path,
                                              f"Models/{model_name}/Clusters={k_val}")

            # Save model and meta data
            try:
                file_path = pickle_object_to_file(model,
                                                  model_path,
                                                  f"{model_name}_Clusters={str(k_val)}")

                pickle_object_to_file(self.__get_centers(model),
                                      model_path,
                                      f"Center points")


                dict_to_json_file({i:i for i in range(0,len(model.get_clusters()))},
                                  model_path,
                                  "Cluster Labels")

                self.__cluster_models_paths[
                    f"{model_name}_Clusters={str(k_val)}"] = file_path

            except:
                print(f"Something went wrong when trying to save the model: {model_name}")

            plt.plot(ks[k_val - 1],
                     inertias_matrix[best_elbow_index].tolist()[0][k_val - 1],
                     'r*')
            best_clusters.append(k_val)

        self.save_plot(f"Models/{model_name}",
                       f"Best of all {model_name} Elbows")

        if display_visuals and self.__notebook_mode:
            plt.show()
        plt.close("all")

        best_clusters.sort()

        if display_visuals and self.__notebook_mode:
            display(proximity_elbow_count)
            display(center_elbow_count)

        return best_clusters

    def __save_update_best_model_clusters(self):
        """

            Generates a directory structure for saving the suggested cluster
            amounts.
        """

        create_dir_structure(self.folder_path,
                             "_Extras")

        pickle_object_to_file(self.__models_suggested_clusters,
                              self.folder_path + "_Extras",
                              "All suggested clusters")
        write_object_text_to_file(self.__models_suggested_clusters,
                                  self.folder_path + "_Extras",
                                  "All suggested clusters")

        all_clusters = []
        for model_name, best_clusters in self.__models_suggested_clusters.items():
            write_object_text_to_file(best_clusters,
                                      self.folder_path + "_Extras",
                                      f"{model_name} suggested clusters")

            all_clusters += best_clusters


        write_object_text_to_file(round(sum(all_clusters) / len(all_clusters)),
                                  self.folder_path + "_Extras",
                                  "Average of suggested clusters")


    def __get_unique_random_indexes(self,
                                    k_val):
        """

            Helper function for pyclustering models that require list indexes
            to the scaled data.

        Args:
            k_val: int
                How many random center points or clusters are required.

        Returns:
            Returns a list of randomly indexes.
        """

        if len(self.__scaled) < k_val:
            raise ValueError("Can't generate more random indexes/center points "
                             "than avaliable.")

        random_indexes = set()
        while len(random_indexes) != k_val:

            index = random.randint(0, len(self.__scaled) - 1)

            if index not in random_indexes:
                random_indexes.add(index)

        return random_indexes


    def __determine_curve(self,
                          intertia):
        """

            When for finding the elbow of the graph it's important to understand
            if the curve of the intertia vector is convex or concaved

        Args:
            intertia: list of floats
                Cluster models evaluated by calculating their intertia.

        Returns:
            Returns a string that says either "concave" or "convex". Returns
            nothing if the intertia vector has a length less than 2.
        """

        # Can't determine concaved or convexed; return None
        if len(intertia) < 2:
            return None

        # Look at the vector in different lengths to determine the best shape
        convex_count = 0
        concave_count = 0
        for i in range(2, len(intertia) + 1):
            x = range(len(intertia))

            warnings.filterwarnings("ignore")

            # Calculate shape
            poly = np.polyfit(x[0:i],
                              intertia[0:i],
                              2)
            warnings.filterwarnings("default")

            # Derivative greater than 0 must be concaved
            if poly[0] >= 0:
                convex_count += 1
            else:
                concave_count += 1

        # Which shape was found more frequently within the vector.
        if concave_count >= convex_count:
            return "concave"
        else:
            return "convex"


    def __display_cluster_label_rank_graph(self,
                                           cluster_labels,
                                           counts,
                                           sub_dir="",
                                           display_visuals=True):
        """

            Darker colors have higher rankings (values)

        Args:
            cluster_labels: list of strings
                List of label names.

            counts: list of ints
                Counts for how many each cluster label had. Order must match
                'cluster_labels'.

            sub_dir: string
                Directory to create and write data to.

            display_visuals: bool
                If set to True will display graphic
        """
        plt.figure(figsize=(12, 7))

        palette = "PuBu"

        # Color ranking
        rank_list = np.argsort(-np.array(counts)).argsort()
        pal = sns.color_palette(palette, len(counts))
        palette = np.array(pal[::-1])[rank_list]

        plt.clf()

        plt.title("Cluster labels and data points count")

        ax = sns.barplot(x=cluster_labels,
                         y=counts,
                         palette=palette,
                         order=cluster_labels)

        # Labels for numerical count of each bar
        for p in ax.patches:
            height = p.get_height()
            ax.text(p.get_x() + p.get_width() / 2.,
                    height + 3,
                    '{:1}'.format(height),
                    ha="center")

        self.save_plot(sub_dir,
                       "Cluster labels and data points count")

        if display_visuals and self.__notebook_mode:
            plt.show()

        plt.close("all")

    def __get_color_name(self,
                         color):
        """

            Simple helper function for a simple switch statement.

        Args:
            color: char
                Single character value representing a real color name.

        Returns:
            A full color name.
        """

        if color == "b":
            return "Blue"

        elif color == "g":
            return "Green"

        elif color == "r":
            return "Red"

        elif color == "c":
            return "Cyan"

        elif color == "m":
            return "Magenta"

        elif color == "y":
            return "Yellow"

        elif color == "k":
            return "Black"

        elif color == "w":
            return "White"

        else:
            return color

    def __nearest_cluster(self,
                          center_points,
                          dp):
        """

            Simple helper function to get the right cluster label. Cluster label
            is determined by how close the given data point is to the center

        Args:
            center_points: list of lists of floats (list of data points)
                Cluster's center point.

            dp: list of floats, (data point)
                Data point
        """

        return np.argmin([euclidean_distance(dp, c)
                          for c in center_points])






# General/Misc somethings that don't quite fit it the other utils
import itertools
import inspect
from eflow.utils.sys_utils import write_object_text_to_file

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


def get_parameters(f):
    """

        Get a the parameters of a given function definition

    Returns:
        Gives back the function's set of parameters
    """
    return set(inspect.getfullargspec(f)[0])


def string_condtional(given_val,
                      full_condtional):
    """

        Returns back a boolean value of whether or not the conditional the
        given value passes the condition.

    Args:
        given_val:
            Numerical value to replace 'x'.

        full_condtional:
            Specified string conditional.
            Ex: x >= 0 and x <=100

    Note:
        Currently only handles and conditionals (no 'or' statements yet)
    """

    try:
        given_val = float(given_val)
    except Exception as _:
        return False

    condtional_returns = []
    operators = [i for i in full_condtional.split(" ")
                 if i == "or" or i == "and"]

    all_condtionals = list(
        itertools.chain(
            *[i.split("or")
              for i in full_condtional.split("and")]))
    for condtional_line in all_condtionals:
        condtional_line = condtional_line.replace(" ", "")
        if condtional_line[0] == 'x':
            condtional_line = condtional_line.replace('x', '')

            condtional = ''.join(
                [i for i in condtional_line if
                 not (i.isdigit() or i == '.')])
            compare_val = float(condtional_line.replace(condtional, ''))
            if condtional == "<=":
                condtional_returns.append(given_val <= compare_val)

            elif condtional == "<":
                condtional_returns.append(given_val < compare_val)

            elif condtional == ">=":
                condtional_returns.append(given_val >= compare_val)

            elif condtional == ">":
                condtional_returns.append(given_val > compare_val)

            elif condtional == "==":
                condtional_returns.append(given_val == compare_val)

            elif condtional == "!=":
                condtional_returns.append(given_val != compare_val)
            else:
                print("ERROR")
                return False
        else:
            print("ERROR")
            return False

    if not len(operators):
        return condtional_returns[0]
    else:
        i = 0
        final_return = None

        for op in operators:
            print(condtional_returns)
            if op == "and":

                if final_return is None:
                    final_return = condtional_returns[i] and \
                                   condtional_returns[i + 1]
                    i += 2
                else:
                    final_return = final_return and condtional_returns[i]
                    i += 1

            else:
                if final_return is None:
                    final_return = condtional_returns[i] or \
                                   condtional_returns[i + 1]
                    i += 2
                else:
                    final_return = final_return or condtional_returns[i]
                    i += 1

        return final_return




# Public python libs
import itertools
import pandas as pd
import numpy as np
from collections import Counter
from collections import defaultdict
import copy
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.model_selection import GridSearchCV
import operator
from dtreeviz.trees import *
from sklearn.preprocessing import LabelEncoder
from matplotlib import pyplot as mp
import os
from IPython.display import display, HTML
import six
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from tqdm import tqdm
from kneed import DataGenerator, KneeLocator
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

import math

class ThreadSafeDict(dict) :
    def __init__(self, * p_arg, ** n_arg) :
        dict.__init__(self, * p_arg, ** n_arg)
        self._lock = threading.Lock()

    def __enter__(self) :
        self._lock.acquire()
        return self

    def __exit__(self, type, value, traceback) :
        self._lock.release()

# --- File specific constants ---
# Import personal "libs"
def enum(**enums):
    return type('Enum', (), enums)


PROJECT = enum(PATH_TO_OUTPUT_FOLDER=''.join(os.getcwd().partition('/eFlow')[0:1]) + "/Figures")


# --- Figures maintaining ---
def check_create_figure_dir(sub_dir):

    directory_pth = PROJECT.PATH_TO_OUTPUT_FOLDER

    for dir in sub_dir.split("/"):
        directory_pth += "/" + dir
        if not os.path.exists(directory_pth):
            os.makedirs(directory_pth)

    return directory_pth


def create_plt_png(sub_dir, filename):

    # Ensure directory structure is init correctly
    abs_path = check_create_figure_dir(sub_dir)

    # Ensure file ext is on the file.
    if filename[-4:] != ".png":
        filename += ".png"

    fig = plt.figure(1)
    fig.savefig(abs_path + "/" + filename, bbox_inches='tight')


def fast_eudis(v1,
               v2):
    dist = [((a - b) ** 2) * w for a, b, w in zip(v1, v2)]
    dist = math.sqrt(sum(dist))
    return dist


def rotate_point(origin,
                 point,
                 angle):
    """
    Rotate a point counterclockwise by a given angle around a given origin.

    The angle should be given in radians.

    # Author link: http://tinyurl.com/y4yz5hco
    """
    ox, oy = origin
    px, py = point
    print(px)

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return qx, qy


# Not created by me!
# Author: http://tinyurl.com/y2hjhbwf
def render_mpl_table(data,
                     sub_dir,
                     filename,
                     col_width=3.0, row_height=0.625, font_size=14,
                     header_color='#40466e', row_colors=['#f1f1f2', 'w'], edge_color='w',
                     bbox=[0, 0, 1, 1], header_columns=0,
                     ax=None, **kwargs):
    if ax is None:
        size = (np.array(data.shape[::-1]) + np.array([0, 1])) * np.array([col_width, row_height])
        fig, ax = plt.subplots(figsize=size)
        ax.axis('off')

    mpl_table = ax.table(cellText=data.values, bbox=bbox, colLabels=data.columns, **kwargs)

    mpl_table.auto_set_font_size(False)
    mpl_table.set_fontsize(font_size)

    for k, cell in six.iteritems(mpl_table._cells):
        cell.set_edgecolor(edge_color)
        if k[0] == 0 or k[1] < header_columns:
            cell.set_text_props(weight='bold', color='w')
            cell.set_facecolor(header_color)
        else:
            cell.set_facecolor(row_colors[k[0]%len(row_colors) ])

    create_plt_png(sub_dir, filename)

    plt.close()


# --- General Utils/Misc ---
def enum(**enums):
    """
        Using past pythons enums b/c I like them better
    """
    return type('Enum', (), enums)


def find_nearest(numbers, target):
    """
        Find the closest fitting number to the target number
    """
    numbers = np.asarray(numbers)
    idx = (np.abs(numbers - target)).argmin()
    return numbers[idx]


def get_max_index_val(given_list):
    """
        Returns the max index and value of a list
    """
    return max(enumerate(given_list), key=operator.itemgetter(1))


# I am this lazy yes...don't hate
def vertical_spacing(spaces=1):
    """
        I could just overwrite the automatic newline builtin.
        But the code looks unclean so screw it.
    """
    for _ in range(0, spaces):
        print()


# --- Pandas dataframe maintaining ---
def replace_df_vals(passed_df, replace_dict):
    """
        Uses a dict of dict to replace dataframe data
    """

    def replace_vals_col(data, decoder):
        return decoder[data]

    df = copy.deepcopy(passed_df)
    for col in df.columns:
        if col in replace_dict.keys():
            df[col] = np.vectorize(replace_vals_col)(
                df[col], replace_dict[col])

    return df


# Returns encoded df and label encoded map
def encode_df(df, df_features):
    
    obj_cols = df_features.categorical_features()

    df = copy.deepcopy(df)
    # ---
    le_map = defaultdict(LabelEncoder)

    # Encoding the variable
    fit = df[obj_cols].apply(lambda x: le_map[x.name].fit_transform(x))

    # Inverse the encoded
    fit.apply(lambda x: le_map[x.name].inverse_transform(x))

    # Using the dictionary to label future data
    df[obj_cols] = df[obj_cols].apply(lambda x: le_map[x.name].transform(x))

    return df, le_map

def decode_df(df,le_map):
    
    df = copy.deepcopy(df)
    decode_cols = list(le_map.keys())
    df[decode_cols] = df[decode_cols].apply(lambda x: le_map[x.name].inverse_transform(x))
    
    return df


def print_encoder(le):

    """
        Display the relationship between encoded values and actual values
    """

    print("\nLabel mapping:\n")
    for i, item in enumerate(le.classes_):
        print("\t", item, '-->', i)

    # Draw a line
    print("-"*30, "\n")

def print_encoder_map(le_map):

    for column_name,le in le_map.items():
        print("\nLabel mapping:\n")

        for i, item in enumerate(le.classes_):
            print("\t", item, '-->', i)

        # Draw a line
        print("-"*30, "\n")


def remove_outliers_df(df, removal_dict):
    """
        Removes outliers with a 'High'/'Low' keyed map;
        Remove all above 'High' and all below 'Low'
    """

    df = copy.deepcopy(df)

    # Loop on columns
    for feature_name in removal_dict.keys():

        # Replacements found; remove all above 'High' and all below 'Low'
        if feature_name in df.columns:
            if removal_dict[feature_name]["High"]:
                df = df[df[feature_name] < removal_dict[feature_name]["High"]]
            elif removal_dict[feature_name]["Low"]:
                df = df[df[feature_name] > removal_dict[feature_name]["Low"]]

    return df.reset_index(drop=True)


def inspect_feature_matrix(matrix,
                           feature_names):
    """
        Creates a dataframe to quickly analyze a matrix
    """
    mean_matrix = np.mean(matrix, axis=0)
    std_matrix = np.std(matrix, axis=0)
    data_dict = dict()
    for index, feature_name in enumerate(feature_names):
        data_dict[feature_name] = [mean_matrix[index],
                                   std_matrix[index]]

    tmp_df = pd.DataFrame.from_dict(data_dict,
                                    orient='index',
                                    columns=['Mean', 'Standard Dev'])

    display(tmp_df)

    return tmp_df


# --- General Graphing ---

# Not created by me!
# Created by my teacher: Narine Hall
def visualize_pca_variance(data):
    """
        Visualize PCA matrix feature importance.
    """

    # Check for pca variance
    pca = PCA()
    pca.fit_transform(data)

    plt.bar(range(1, len(pca.explained_variance_ratio_)+1), pca.explained_variance_ratio_)
    plt.xticks()
    plt.ylabel('variance ratio')
    plt.xlabel('PCA feature')
    plt.tight_layout()
    plt.savefig('figures/pca_variance_ratio.png')

    plt.bar(range(1, len(pca.explained_variance_ratio_)+1), pca.explained_variance_ratio_.cumsum())
    plt.xticks()
    plt.ylabel('cumulative sum of variances')
    plt.xlabel('PCA feature')
    plt.tight_layout()
    create_plt_png("Default/",
                   "Rank_Graph_")

    return pca

def plot_confusion_matrix(cm,
                          classes,
                          normalize=False,
                          title='Confusion matrix',
                          cmap=plt.cm.Blues):
    plt.imshow(cm,
               interpolation='nearest',
               cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, cm[i, j],
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')



def display_rank_graph(feature_names, metric,
                       title="", y_title="",
                       x_title="",
                       plt_output="Default/Analysis"):
    """
        Darker colors that have higher rankings (values)
    """
    plt.figure(figsize=(7, 7))

    # Init color ranking fo plot
    # Ref: http://tinyurl.com/ydgjtmty
    pal = sns.color_palette("GnBu_d", len(metric))
    rank = np.array(metric).argsort().argsort()
    ax = sns.barplot(y=feature_names, x=metric,
                     palette=np.array(pal[::-1])[rank])
    plt.xticks(rotation=0, fontsize=15)
    plt.yticks(fontsize=15)
    plt.xlabel(x_title, fontsize=20, labelpad=20)
    plt.ylabel(y_title, fontsize=20, labelpad=20)
    plt.title(title, fontsize=15)

    plt.show()
    plt.close()


# --- Some sklearn specific functions ---
def optimize_model_grid(model,
                        X_train,
                        y_train,
                        param_grid,
                        n_jobs,
                        cv=10):
    """
        Finds the best parameters for a grid; returns the model and parameters.
    """
    # Instantiate the GridSearchCV object
    model_cv = GridSearchCV(model, param_grid, cv=cv, n_jobs=n_jobs)

    # Fit it to the data
    model_cv.fit(X_train, y_train)

    # Print the tuned parameters and score
    print("Tuned Parameters: {}".format(model_cv.best_params_))
    print("Best score on trained data was {0:4f}".format(model_cv.best_score_))

    model = type(model)(**model_cv.best_params_)

    # Return model and parameters
    return model, model_cv.best_params_


# Not created by me!
# Author: https://github.com/scikit-learn/scikit-learn/issues/7845
def report_to_dict(cr):

    """
        Converts a 'classification_report' to a dataframe for more professional look.
    """

    # Parse rows
    tmp = list()
    for row in cr.split("\n"):
        parsed_row = [x for x in row.split("  ") if len(x) > 0]
        if len(parsed_row) > 0:
            tmp.append(parsed_row)

    # Store in dictionary
    measures = tmp[0]

    D_class_data = defaultdict(dict)
    for row in tmp[1:]:
        class_label = row[0]
        for j, m in enumerate(measures):
            D_class_data[class_label][m.strip()] = float(row[j + 1].strip())
    return D_class_data


def create_decorrelate_df(df,
                          df_features,
                          target_name,
                          desired_col_average=0.5,
                          show=True):
    df = copy.deepcopy(df)
    df_features = copy.deepcopy(df_features)
    while True:

        # Display correlation map
        corr_metrics = df.corr()
        if show and PROJECT.IS_NOTEBOOK:
            display(corr_metrics.style.background_gradient())

        # Get the correlation means of each feature
        corr_feature_means = []
        for feature_name in list(corr_metrics.columns):

            # Ignore target feature; Only a problem if target was numerical
            if target_name != feature_name:
                corr_feature_means.append(corr_metrics[feature_name].mean())

        if show:
            # Display graph rank
            display_rank_graph(feature_names=list(corr_metrics.columns),
                               metric=corr_feature_means,
                               title="Average Feature Correlation",
                               y_title="Correlation Average",
                               x_title="Features")

        index, max_val = get_max_index_val(corr_feature_means)

        if max_val > desired_col_average:
            # Drop col and notify
            feature_name = list(corr_metrics.columns)[index]
            df.drop(feature_name, axis=1, inplace=True)
            df_features.remove(feature_name)
            print("Dropped column: {0}".format(feature_name))
            vertical_spacing(5)

        # End loop desired average reached
        else:
            if show and PROJECT.IS_NOTEBOOK:
                display(corr_feature_means)
            break

    return df, df_features


def random_partition_of_random_samples(list_of_df_indexes,
                                       random_sampled_rows,
                                       random_sample_amount,
                                       random_state=None):
    # Convert to numpy array if list
    if isinstance(list_of_df_indexes, list):
        list_of_df_indexes = np.array(list_of_df_indexes)

    np.random.seed(random_state)
    for _ in range(np.random.randint(1, 3)):
        np.random.shuffle(list_of_df_indexes)

    if random_sample_amount > len(list_of_df_indexes):
        random_sample_amount = len(list_of_df_indexes)

    return_matrix = np.zeros((random_sample_amount, random_sampled_rows))
    for i in range(random_sampled_rows):
        sub_list = list_of_df_indexes[:random_sample_amount]
        return_matrix[i] = sub_list
        np.random.shuffle(list_of_df_indexes)

    np.random.seed(None)
    return return_matrix


def kmeans_impurity_sample_removal(df,
                                   target,
                                   pca_perc,
                                   majority_class,
                                   majority_class_threshold=.5,
                                   random_state=None):
    """
        Generate models based on the found 'elbow' of the interia values.
    """
    df = copy.deepcopy(df)
    removal_df_indexes = []

    # Create scaler object
    scaler = StandardScaler()
    scaled = scaler.fit_transform(df.drop(columns=[target]))

    print("\nInspecting scaled results!")
    # self.__inspect_feature_matrix(matrix=scaled,
    #                               feature_names=df.columns)

    pca = PCA()
    scaled = pca.fit_transform(scaled)

    # Generate "dummy" feature names
    pca_feature_names = ["PCA_Feature_" +
                         str(i) for i in range(1,
                                               len(df.columns) + 1)]

    print("\nInspecting applied pca results!")
    # self.__inspect_feature_matrix(matrix=scaled,
    #                               feature_names=pca_feature_names)

    if pca_perc < 1.0:
        # Find cut off point on cumulative sum
        cutoff_index = np.where(
            pca.explained_variance_ratio_.cumsum() > pca_perc)[0][0]
    else:
        cutoff_index = scaled.shape[1] - 1

    print(
        "After applying pca with a cutoff percentage {0}%"
        " for the cumulative index. Using features 1 to {1}".format(
            pca_perc, cutoff_index + 1))

    print("Old shape {0}".format(scaled.shape))

    scaled = scaled[:, :cutoff_index + 1]
    pca_feature_names = pca_feature_names[0: cutoff_index + 1]

    print("New shape {0}".format(scaled.shape))

    scaled = scaler.fit_transform(scaled)

    print("\nInspecting re-applied scaled results!")
    # self.__inspect_feature_matrix(matrix=scaled,
    #                               feature_names=pca_feature_names)

    ks = range(1, 15)
    inertias = []
    all_models = []

    for k in tqdm(ks):
        # Create a KMeans instance with k clusters: model
        model = KMeans(n_clusters=k,
                       random_state=random_state).fit(scaled)

        # Append the inertia to the list of inertias
        inertias.append(model.inertia_)
        all_models.append(model)

    a = KneeLocator(inertias, ks, curve='convex', direction='decreasing')
    elbow_index = np.where(inertias == a.knee)
    best_worst_model_index = elbow_index[0][0] + 2
    best_worst_model = None

    if len(all_models) < best_worst_model_index:
        print("OUT OF INDEX ERROR!!!")
    else:
        best_worst_model = all_models[best_worst_model_index]

    print(best_worst_model.labels_)
    df["Cluster_Name"] = model.labels_
    for val in set(df["Cluster_Name"]):
        sub_df = df[df["Cluster_Name"] == val]
        val_counts_dict = sub_df[target].value_counts().to_dict()

        if majority_class in val_counts_dict and val_counts_dict[
            majority_class] / sum(val_counts_dict.values()) <= majority_class_threshold:
            removal_df_indexes += sub_df[
                sub_df[target] == majority_class].index.values.tolist()

    return removal_df_indexes


import nltk
from nltk.corpus import wordnet


def get_synonyms_antonyms(word):

    synonyms = []
    antonyms = []

    for syn in wordnet.synsets(word):
        for l in syn.lemmas():
            synonyms.append(l.name())
            if l.antonyms():
                antonyms.append(l.antonyms()[0].name())

    return {word.replace("_", " ") for word in synonyms}, \
           {word.replace("_", " ") for word in antonyms}


def get_synonyms(word):
    synonyms = []

    for syn in wordnet.synsets(word):
        for l in syn.lemmas():
            synonyms.append(l.name())

    return {word.replace("_", " ") for word in synonyms}

def get_antonyms(word):
    antonyms = []

    for syn in wordnet.synsets(word):
        for l in syn.lemmas():
            if l.antonyms():
                antonyms.append(l.antonyms()[0].name())

    return {word.replace("_", " ") for word in antonyms}